# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import os
import ast
import xml.etree.ElementTree as ET
import numpy as np
import logging
import pint
import traits.api as t

from hyperspy.misc.io.tools import convert_xml_to_dict


_logger = logging.getLogger(__name__)
_ureg = pint.UnitRegistry()


# Plugin characteristics
# ----------------------
format_name = 'empad'
description = ''
full_support = False
# Recognised file extension
file_extensions = ['xml', 'XML']
default_extension = 0

 # Writing capabilities:
writes = False


def _read_raw(info, fp, lazy=False):

    raw_height = info['raw_height']
    width = info['width']
    height = info['height']

    if lazy:
        data = np.memmap(fp, dtype='<f4', mode='r')
    else:
        data = np.fromfile(fp, dtype='<f4')

    if 'series_count' in info.keys():   # stack of images
        size = (info['series_count'], raw_height, width)
        data = data.reshape(size)[..., :height, :]

    else:  # 2D x 2D
        size = (info['scan_x'], info['scan_y'], raw_height, width)
        data = data.reshape(size)[..., :height, :]
    return data


def _parse_xml(filename):
    tree = ET.parse(filename)
    om = convert_xml_to_dict(tree.getroot())

    info = {'raw_filename': om.root.raw_file.filename,
            'width':128,
            'height':128,
            'raw_height':130,
            'record-by':'image'}
    if om.has_item('root.count'):
        # Stack of images
        info.update({'series_count':int(om.root.scan_parameters.series_count)})
    elif om.has_item('root.pix_x') and om.has_item('root.pix_y'):
        # 2D x 2D
        info.update({'scan_x':int(om.root.pix_x),
                     'scan_y':int(om.root.pix_y)})
    else:
        raise IOError("Unsupported Empad file: the scan parameters cannot "
                      "be imported.")

    return om, info


def _convert_scale_units(value, units, factor=1):
    v = np.float(value) * _ureg(units)
    converted_v = (factor * v).to_compact()
    converted_value = converted_v.magnitude / factor
    converted_units = '{:~}'.format(converted_v.units)

    return converted_value, converted_units


def file_reader(filename, lazy=False, **kwds):

    om, info = _parse_xml(filename)
    dname, fname = os.path.split(filename)

    md = {
        'General': {'original_filename': fname,
                    'title': os.path.splitext(fname)[0],
                    },
        "Signal": {'signal_type': 'electron_diffraction',
                   'record_by': 'image'},
    }

    if om.has_item('root.timestamp.isoformat'):
        date, time = om.root.timestamp.isoformat.split('T')
        md['General'].update({"date":date, "time":time})

    units = [t.Undefined, t.Undefined]
    scales = [1, 1]
    origins = [-64, -64]
    axes = []
    index_in_array = 0
    names = ['height', 'width']

    if 'series_count' in info.keys():
        names = ['series_count'] + names
        units.insert(0, 'ms')
        scales.insert(0, 1)
        origins.insert(0, 0)
    else:
        names = ['scan_x', 'scan_y'] + names
        units.insert(0, '')
        units.insert(0, '')
        scales.insert(0, 1)
        scales.insert(0, 1)
        origins.insert(0, 0)
        origins.insert(0, 0)

    sizes = [info[name] for name in names]

    if not 'series_count' in info.keys():
        try:
            fov = ast.literal_eval(
                om.root.iom_measurements.opticsget_full_scan_field_of_view)
            for i in range(2):
                value = fov[i] / sizes[i]
                scales[i], units[i] = _convert_scale_units(value, 'm', sizes[i])
        except BaseException:
            _logger.warning("The scale of the navigation axes cannot be read.")

    try:
        pixel_size = float(om.root.iom_measurements.calibrated_pixelsize) * 1E9
        for i in [-1, -2]:
            scales[i] = pixel_size
            units[i] = '1/nm'
    except BaseException:
        _logger.warning("The scale of the signal axes cannot be read.")

    for i in range(len(names)):
        if sizes[i] > 1:
            axes.append({
                'size': sizes[i],
                'index_in_array': index_in_array,
                'name': names[i],
                'scale': scales[i],
                'offset': origins[i] * scales[i],
                'units': units[i],
            })
            index_in_array += 1

    data = _read_raw(info,
                     os.path.join(dname, info['raw_filename']),
                     lazy=lazy)

    dictionary = {
        'data': data.squeeze(),
        'axes': axes,
        'metadata': md,
        'original_metadata': om.as_dictionary()
    }

    return [dictionary, ]



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


from unittest import mock

import numpy as np
import pytest

from hyperspy.component import Parameter


class Dummy:

    def __init__(self):
        self.value = 1

    def add_one(self):
        self.value += 1


class DummyAxesManager:
    navigation_shape = ()
    indices = ()

    @property
    def _navigation_shape_in_array(self):
        return self.navigation_shape[::-1]


class TestParameterLen1:

    def setup_method(self, method):
        self.par = Parameter()

    def test_set_value(self):
        self.par.value = 2
        assert self.par.value == 2

    def test_set_value_wrong_length(self):
        with pytest.raises(ValueError):
            self.par.value = (2, 2)

    def test_set_value_bounded(self):
        self.par.bmin = 1
        self.par.bmax = 3
        self.par.ext_bounded = True
        assert self.par.value == 1
        self.par.value = 1.5
        assert self.par.value == 1.5
        self.par.value = 0.5
        assert self.par.value == 1
        self.par.value = 4
        assert self.par.value == 3
        self.par.bmax = 2
        assert self.par.value == 2
        self.par.value = 1.5
        self.par.bmin = 1.6
        assert self.par.value == 1.6

    def test_ext_force_positive(self):
        self.par.ext_bounded = True
        self.par.ext_force_positive = True
        self.par.value = -1.5
        assert self.par.value == 1.5
        self.par.bmax = 2
        self.par.value = -3
        assert self.par.value == 2

    def test_number_of_elements(self):
        assert len(self.par) == 1

    def test_default_value(self):
        assert self.par.value == 0

    def test_connect_disconnect(self):
        dummy = Dummy()
        self.par.events.value_changed.connect(dummy.add_one, [])
        self.par.value = 1
        assert dummy.value == 2

        # Setting the same value should not call the connected functions
        self.par.value = 1
        assert dummy.value == 2

        # After disconnecting dummy.value should not change
        self.par.events.value_changed.disconnect(dummy.add_one)
        self.par.value = 2
        assert dummy.value == 2

    def test_map_size0(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._create_array()
        self.par.value = 1
        self.par.std = 0.1
        self.par.store_current_value_in_array()
        assert self.par.map['values'][0] == 1
        assert self.par.map['is_set'][0]
        assert self.par.map['std'][0] == 0.1

    def test_map_size1(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._axes_manager.navigation_shape = [1, ]
        self.par._create_array()
        self.par.value = 1
        self.par.std = 0.1
        self.par.store_current_value_in_array()
        assert self.par.map['values'][0] == 1
        assert self.par.map['is_set'][0]
        assert self.par.map['std'][0] == 0.1

    def test_map_size2(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._axes_manager.navigation_shape = [2, ]
        self.par._axes_manager.indices = (1,)
        self.par._create_array()
        self.par.value = 1
        self.par.std = 0.1
        self.par.store_current_value_in_array()
        assert self.par.map['values'][1] == 1
        assert self.par.map['is_set'][1]
        assert self.par.map['std'][1] == 0.1


class TestParameterLen2:

    def setup_method(self, method):
        self.par = Parameter()
        self.par._number_of_elements = 2

    def test_set_value(self):
        self.par.value = (2, 2)
        assert self.par.value == (2, 2)

    def test_set_value_wrong_length(self):
        with pytest.raises(ValueError):
            self.par.value = 2

    def test_set_value_wrong_length2(self):
        with pytest.raises(ValueError):
            self.par.value = (2, 2, 2)

    def test_set_value_bounded(self):
        self.par.bmin = 1
        self.par.bmax = 3
        self.par.ext_bounded = True
        self.par.value = (1.5, 1.5)
        assert self.par.value == (1.5, 1.5)
        self.par.value = (0.5, 0.5)
        assert self.par.value == (1, 1)
        self.par.value = (4, 4)
        assert self.par.value == (3, 3)

    def test_ext_force_positive(self):
        self.par.ext_bounded = True
        self.par.ext_force_positive = True
        self.par.value = (2, -1.5)
        assert self.par.value == (2, 1.5)

    def test_number_of_elements(self):
        assert len(self.par) == 2

    def test_default_value(self):
        assert self.par.value == (0, 0)

    def test_connect_disconnect(self):
        dummy = Dummy()
        self.par.events.value_changed.connect(dummy.add_one, [])
        self.par.value = (1, 1)
        assert dummy.value == 2

        # Setting the same value should not call the connected functions
        self.par.value = (1, 1)
        assert dummy.value == 2

        # After disconnecting dummy.value should not change
        self.par.events.value_changed.disconnect(dummy.add_one)
        self.par.value = (2, 2)
        assert dummy.value == 2

    def test_map_size0(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._create_array()
        self.par.value = (1, 1)
        self.par.std = (0.1, 0.1)
        self.par.store_current_value_in_array()
        assert tuple(self.par.map['values'][0]) == (1, 1)
        assert self.par.map['is_set'][0]
        assert tuple(self.par.map['std'][0]) == (0.1, 0.1)

    def test_map_size1(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._axes_manager.navigation_shape = [1, ]
        self.par._create_array()
        self.par.value = (1, 1)
        self.par.std = (0.1, 0.1)
        self.par.store_current_value_in_array()
        assert tuple(self.par.map['values'][0]) == (1, 1)
        assert self.par.map['is_set'][0]
        assert tuple(self.par.map['std'][0]) == (0.1, 0.1)

    def test_map_size2(self):
        self.par._axes_manager = DummyAxesManager()
        self.par._axes_manager.navigation_shape = [2, ]
        self.par._axes_manager.indices = (1,)
        self.par._create_array()
        self.par.value = (1, 1)
        self.par.std = (0.1, 0.1)
        self.par.store_current_value_in_array()
        assert tuple(self.par.map['values'][1]) == (1, 1)
        assert self.par.map['is_set'][1]
        assert tuple(self.par.map['std'][1]) == (0.1, 0.1)

    def test_is_tuple(self):
        self.par.value = np.array((1, 2))
        assert isinstance(self.par.value, tuple)
        self.par.value = [1, 2]
        assert isinstance(self.par.value, tuple)


class TestParameterTwin:

    def setup_method(self, method):
        self.p1 = Parameter()
        self.p2 = Parameter()

    def test_slave_fixed(self):
        assert self.p2.free
        self.p2.twin = self.p1
        assert not self.p2.free

    def test_twin_value(self):
        self.p2.twin = self.p1
        self.p1.value = 3
        assert self.p1.value == self.p2.value
        self.p2.twin = None
        assert self.p1.value == self.p2.value
        self.p1.value = 2
        assert 3 == self.p2.value

    def test_twin_value_bounded(self):
        self.p2.bmax = 2
        self.p2.ext_bounded = True
        self.p2.twin = self.p1
        self.p1.value = 3
        assert self.p1.value == self.p2.value
        self.p2.twin = None
        assert self.p2.value == self.p2.bmax

    def test_twin_function(self):
        self.p2.twin_function = lambda x: x + 2
        self.p2.twin_inverse_function = lambda x: x - 2
        self.p2.twin = self.p1
        assert self.p1.value == self.p2.value - 2
        self.p2.value = 10
        assert self.p1.value == 8

    def test_inherit_connections(self):
        dummy = Dummy()
        self.p2.events.value_changed.connect(dummy.add_one, [])
        self.p2.twin = self.p1
        self.p1.value = 2
        assert dummy.value == 2
        # The next line calls add_one -> value = 3
        self.p2.twin = None
        # Next one shouldn't call add_one -> value = 3
        self.p1.value = 4
        assert dummy.value == 3
        self.p2.value = 10
        assert dummy.value == 4


class TestGeneralMethods:

    def setup_method(self, method):
        self.par = Parameter()
        self.par._axes_manager = mock.MagicMock()
        self.par.map = np.array(
            [(a, b, c) for a, b, c in zip([1, 3, 5], [2, 4, 6], [0, 0, 0])],
            dtype=[('values', 'float'), ('std', 'float'), ('is_set', bool)])

    def test_as_signal(self):
        par = self.par

        # additional setup
        par._axes_manager._get_navigation_axes_dicts.return_value = [
            {'name': 'one', 'navigate': True,
             'offset': 0.0, 'scale':
             1.0, 'size': 3,
             'units': 'bar'}, ]
        par._number_of_elements = 2
        par.component = mock.MagicMock()
        par.component.active_is_multidimensional = True
        par.component._active_array = np.array([1, 0, 1], dtype=bool)

        # testing
        s = par.as_signal('std')
        np.testing.assert_array_equal(s.data, np.array([2, np.nan, 6]))
        assert s.axes_manager[-1].name == 'one'
        assert par._axes_manager._get_navigation_axes_dicts.called
        assert len(s.axes_manager._axes) == 2
        assert not s.axes_manager[-1].navigate
        assert s.axes_manager[0].navigate
        assert s.axes_manager[0].size == 2

    def test_store_current_values_normal_indices(self):
        par = self.par
        par._axes_manager.indices = (1,)
        par.value = 3.5
        par.std = 4.5
        par.store_current_value_in_array()
        assert par.map['is_set'][1]
        assert par.map['std'][1] == 4.5
        assert par.map['values'][1] == 3.5

    def test_store_current_values_no_indices(self):
        par = self.par
        par._axes_manager.indices = ()
        par.value = 3.5
        par.std = 4.5
        par.store_current_value_in_array()
        assert par.map['is_set'][0]
        assert par.map['std'][0] == 4.5
        assert par.map['values'][0] == 3.5



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging

_logger = logging.getLogger(__name__)


from hyperspy import docstrings

__doc__ = """
HyperSpy: a multi-dimensional data analysis package for Python
==============================================================

Documentation is available in the docstrings and online at
http://hyperspy.org/hyperspy-doc/current/index.html.

All public packages, functions and classes are in :mod:`~hyperspy.api`. All
other packages and modules are for internal consumption and should not be
needed for data analysis.

%s

More details in the :mod:`~hyperspy.api` docstring.

""" % docstrings.START_HSPY


from . import Release

__all__ = ["api"]
__version__ = Release.version



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest

from hyperspy.misc.machine_learning.import_sklearn import sklearn_installed
from hyperspy.signals import Signal1D


def test_learning_results_decom():
    rng = np.random.RandomState(123)

    s1 = Signal1D(rng.random_sample(size=(20, 100)))
    s1.decomposition(output_dimension=2)

    out = str(s1.learning_results)
    assert "Decomposition parameters" in out
    assert "algorithm=SVD" in out
    assert "output_dimension=2" in out
    assert "Demixing parameters" not in out


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_learning_results_bss():
    rng = np.random.RandomState(123)

    s1 = Signal1D(rng.random_sample(size=(20, 100)))
    s1.decomposition(output_dimension=2)
    s1.blind_source_separation(number_of_components=2)

    out = str(s1.learning_results)
    assert "Decomposition parameters" in out
    assert "Demixing parameters" in out
    assert "algorithm=sklearn_fastica" in out
    assert "n_components=2" in out



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import dask.array as da
import numpy as np

from hyperspy.signals import Signal1D, Signal2D


def test_lazy_changetype_rechunk():
    ar = da.ones((50, 50, 256, 256), chunks=(5, 5, 256, 256), dtype="uint8")
    s = Signal2D(ar).as_lazy()
    s._make_lazy(rechunk=True)
    assert s.data.dtype is np.dtype("uint8")
    chunks_old = s.data.chunks
    s.change_dtype("float")
    assert s.data.dtype is np.dtype("float")
    chunks_new = s.data.chunks
    assert len(chunks_old[0]) * len(chunks_old[1]) < len(chunks_new[0]) * len(
        chunks_new[1]
    )
    s.change_dtype("uint8")
    assert s.data.dtype is np.dtype("uint8")
    chunks_newest = s.data.chunks
    assert chunks_newest == chunks_new


def test_lazy_changetype_rechunk_False():
    ar = da.ones((50, 50, 256, 256), chunks=(5, 5, 256, 256), dtype="uint8")
    s = Signal2D(ar).as_lazy()
    s._make_lazy(rechunk=True)
    assert s.data.dtype is np.dtype("uint8")
    chunks_old = s.data.chunks
    s.change_dtype("float", rechunk=False)
    assert s.data.dtype is np.dtype("float")
    assert chunks_old == s.data.chunks


def test_lazy_reduce_rechunk():
    s = Signal1D(da.ones((10, 100), chunks=(1, 2))).as_lazy()
    reduce_methods = (
        s.sum,
        s.mean,
        s.max,
        s.std,
        s.var,
        s.nansum,
        s.nanmax,
        s.nanmin,
        s.nanmean,
        s.nanstd,
        s.nanvar,
        s.indexmin,
        s.indexmax,
        s.valuemax,
        s.valuemin,
    )
    for rm in reduce_methods:
        assert rm(axis=0).data.chunks == ((100,),)  # The data has been rechunked
        assert rm(axis=0, rechunk=False).data.chunks == (
            (2,) * 50,
        )  # The data has not been rechunked


def test_lazy_diff_rechunk():
    s = Signal1D(da.ones((10, 100), chunks=(1, 2))).as_lazy()
    for rm in (s.derivative, s.diff):
        # The data has been rechunked
        assert rm(axis=-1).data.chunks == ((10,), (99,))
        assert rm(axis=-1, rechunk=False).data.chunks == (
            (1,) * 10,
            (1,) * 99,
        )  # The data has not been rechunked



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
from functools import partial
import warnings

import numpy as np
import dask.array as da
import dask.delayed as dd
from dask import threaded
from dask.diagnostics import ProgressBar
from itertools import product


from hyperspy.signal import BaseSignal
from hyperspy.defaults_parser import preferences
from hyperspy.docstrings.signal import SHOW_PROGRESSBAR_ARG
from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.external.progressbar import progressbar
from hyperspy.misc.array_tools import _requires_linear_rebin
from hyperspy.misc.hist_tools import histogram_dask
from hyperspy.misc.machine_learning import import_sklearn
from hyperspy.misc.utils import multiply, dummy_context_manager

_logger = logging.getLogger(__name__)

lazyerror = NotImplementedError('This method is not available in lazy signals')


def to_array(thing, chunks=None):
    """Accepts BaseSignal, dask or numpy arrays and always produces either
    numpy or dask array.

    Parameters
    ----------
    thing : {BaseSignal, dask.array.Array, numpy.ndarray}
        the thing to be converted
    chunks : {None, tuple of tuples}
        If None, the returned value is a numpy array. Otherwise returns dask
        array with the chunks as specified.

    Returns
    -------
    res : {numpy.ndarray, dask.array.Array}
    """
    if thing is None:
        return None
    if isinstance(thing, BaseSignal):
        thing = thing.data
    if chunks is None:
        if isinstance(thing, da.Array):
            thing = thing.compute()
        if isinstance(thing, np.ndarray):
            return thing
        else:
            raise ValueError
    else:
        if isinstance(thing, np.ndarray):
            thing = da.from_array(thing, chunks=chunks)
        if isinstance(thing, da.Array):
            if thing.chunks != chunks:
                thing = thing.rechunk(chunks)
            return thing
        else:
            raise ValueError


class LazySignal(BaseSignal):
    """A Lazy Signal instance that delays computation until explicitly saved
    (assuming storing the full result of computation in memory is not feasible)
    """
    _lazy = True

    def compute(self, close_file=False, show_progressbar=None, **kwargs):
        """Attempt to store the full signal in memory.

        Parameters
        ----------
        close_file : bool, default False
            If True, attemp to close the file associated with the dask
            array data if any. Note that closing the file will make all other
            associated lazy signals inoperative.
        %s

        Returns
        -------
        None

        """
        if "progressbar" in kwargs:
            warnings.warn(
                "The `progressbar` keyword is deprecated and will be removed "
                "in HyperSpy 2.0. Use `show_progressbar` instead.",
                VisibleDeprecationWarning,
            )
            show_progressbar = kwargs["progressbar"]

        if show_progressbar is None:
            show_progressbar = preferences.General.show_progressbar

        cm = ProgressBar if show_progressbar else dummy_context_manager

        with cm():
            da = self.data
            data = da.compute()
            if close_file:
                self.close_file()
            self.data = data

        self._lazy = False
        self._assign_subclass()

    compute.__doc__ %= SHOW_PROGRESSBAR_ARG

    def close_file(self):
        """Closes the associated data file if any.

        Currently it only supports closing the file associated with a dask
        array created from an h5py DataSet (default HyperSpy hdf5 reader).

        """
        arrkey = None
        for key in self.data.dask.keys():
            if "array-original" in key:
                arrkey = key
                break
        if arrkey:
            try:
                self.data.dask[arrkey].file.close()
            except AttributeError:
                _logger.exception("Failed to close lazy Signal file")

    def _get_dask_chunks(self, axis=None, dtype=None):
        """Returns dask chunks.

        Aims:
            - Have at least one signal (or specified axis) in a single chunk,
              or as many as fit in memory

        Parameters
        ----------
        axis : {int, string, None, axis, tuple}
            If axis is None (default), returns chunks for current data shape so
            that at least one signal is in the chunk. If an axis is specified,
            only that particular axis is guaranteed to be "not sliced".
        dtype : {string, np.dtype}
            The dtype of target chunks.

        Returns
        -------
        Tuple of tuples, dask chunks
        """
        dc = self.data
        dcshape = dc.shape
        for _axis in self.axes_manager._axes:
            if _axis.index_in_array < len(dcshape):
                _axis.size = int(dcshape[_axis.index_in_array])

        if axis is not None:
            need_axes = self.axes_manager[axis]
            if not np.iterable(need_axes):
                need_axes = [need_axes, ]
        else:
            need_axes = self.axes_manager.signal_axes

        if dtype is None:
            dtype = dc.dtype
        elif not isinstance(dtype, np.dtype):
            dtype = np.dtype(dtype)
        typesize = max(dtype.itemsize, dc.dtype.itemsize)
        want_to_keep = multiply([ax.size for ax in need_axes]) * typesize

        # @mrocklin reccomends to have around 100MB chunks, so we do that:
        num_that_fit = int(100. * 2.**20 / want_to_keep)

        # want to have at least one "signal" per chunk
        if num_that_fit < 2:
            chunks = [tuple(1 for _ in range(i)) for i in dc.shape]
            for ax in need_axes:
                chunks[ax.index_in_array] = dc.shape[ax.index_in_array],
            return tuple(chunks)

        sizes = [
            ax.size for ax in self.axes_manager._axes if ax not in need_axes
        ]
        indices = [
            ax.index_in_array for ax in self.axes_manager._axes
            if ax not in need_axes
        ]

        while True:
            if multiply(sizes) <= num_that_fit:
                break

            i = np.argmax(sizes)
            sizes[i] = np.floor(sizes[i] / 2)
        chunks = []
        ndim = len(dc.shape)
        for i in range(ndim):
            if i in indices:
                size = float(dc.shape[i])
                split_array = np.array_split(
                    np.arange(size), np.ceil(size / sizes[indices.index(i)]))
                chunks.append(tuple(len(sp) for sp in split_array))
            else:
                chunks.append((dc.shape[i], ))
        return tuple(chunks)

    def _make_lazy(self, axis=None, rechunk=False, dtype=None):
        self.data = self._lazy_data(axis=axis, rechunk=rechunk, dtype=dtype)

    def change_dtype(self, dtype, rechunk=True):
        from hyperspy.misc import rgb_tools
        if not isinstance(dtype, np.dtype) and (dtype not in
                                                rgb_tools.rgb_dtypes):
            dtype = np.dtype(dtype)
            self._make_lazy(rechunk=rechunk, dtype=dtype)
        super().change_dtype(dtype)
    change_dtype.__doc__ = BaseSignal.change_dtype.__doc__

    def _lazy_data(self, axis=None, rechunk=True, dtype=None):
        """Return the data as a dask array, rechunked if necessary.

        Parameters
        ----------
        axis: None, DataAxis or tuple of data axes
            The data axis that must not be broken into chunks when `rechunk`
            is `True`. If None, it defaults to the current signal axes.
        rechunk: bool, "dask_auto"
            If `True`, it rechunks the data if necessary making sure that the
            axes in ``axis`` are not split into chunks. If `False` it does
            not rechunk at least the data is not a dask array, in which case
            it chunks as if rechunk was `True`. If "dask_auto", rechunk if
            necessary using dask's automatic chunk guessing.

        """
        if rechunk == "dask_auto":
            new_chunks = "auto"
        else:
            new_chunks = self._get_dask_chunks(axis=axis, dtype=dtype)
        if isinstance(self.data, da.Array):
            res = self.data
            if self.data.chunks != new_chunks and rechunk:
                _logger.info(
                    "Rechunking.\nOriginal chunks: %s" % str(self.data.chunks))
                res = self.data.rechunk(new_chunks)
                _logger.info(
                    "Final chunks: %s " % str(res.chunks))
        else:
            if isinstance(self.data, np.ma.masked_array):
                data = np.where(self.data.mask, np.nan, self.data)
            else:
                data = self.data
            res = da.from_array(data, chunks=new_chunks)
        assert isinstance(res, da.Array)
        return res

    def _apply_function_on_data_and_remove_axis(self, function, axes,
                                                out=None, rechunk=True):
        def get_dask_function(numpy_name):
            # Translate from the default numpy to dask functions
            translations = {'amax': 'max', 'amin': 'min'}
            if numpy_name in translations:
                numpy_name = translations[numpy_name]
            return getattr(da, numpy_name)

        function = get_dask_function(function.__name__)
        axes = self.axes_manager[axes]
        if not np.iterable(axes):
            axes = (axes, )
        ar_axes = tuple(ax.index_in_array for ax in axes)
        if len(ar_axes) == 1:
            ar_axes = ar_axes[0]
        # For reduce operations the actual signal and navigation
        # axes configuration does not matter. Hence we leave
        # dask guess the chunks
        if rechunk is True:
            rechunk = "dask_auto"
        current_data = self._lazy_data(rechunk=rechunk)
        # Apply reducing function
        new_data = function(current_data, axis=ar_axes)
        if not new_data.ndim:
            new_data = new_data.reshape((1, ))
        if out:
            if out.data.shape == new_data.shape:
                out.data = new_data
                out.events.data_changed.trigger(obj=out)
            else:
                raise ValueError(
                    "The output shape %s does not match  the shape of "
                    "`out` %s" % (new_data.shape, out.data.shape))
        else:
            s = self._deepcopy_with_new_data(new_data)
            s._remove_axis([ax.index_in_axes_manager for ax in axes])
            return s

    def rebin(self, new_shape=None, scale=None,
              crop=False, out=None, rechunk=True):
        factors = self._validate_rebin_args_and_get_factors(
            new_shape=new_shape,
            scale=scale)
        if _requires_linear_rebin(arr=self.data, scale=factors):
            if new_shape:
                raise NotImplementedError(
                    "Lazy rebin requires that the new shape is a divisor "
                    "of the original signal shape e.g. if original shape "
                    "(10| 6), new_shape=(5| 3) is valid, (3 | 4) is not.")
            else:
                raise NotImplementedError(
                    "Lazy rebin requires scale to be integer and divisor of the "
                    "original signal shape")
        axis = {ax.index_in_array: ax
                for ax in self.axes_manager._axes}[factors.argmax()]
        self._make_lazy(axis=axis, rechunk=rechunk)
        return super().rebin(new_shape=new_shape,
                             scale=scale, crop=crop, out=out)
    rebin.__doc__ = BaseSignal.rebin.__doc__

    def __array__(self, dtype=None):
        return self.data.__array__(dtype=dtype)

    def _make_sure_data_is_contiguous(self):
        self._make_lazy(rechunk=True)

    def diff(self, axis, order=1, out=None, rechunk=True):
        arr_axis = self.axes_manager[axis].index_in_array

        def dask_diff(arr, n, axis):
            # assume arr is da.Array already
            n = int(n)
            if n == 0:
                return arr
            if n < 0:
                raise ValueError("order must be positive")
            nd = len(arr.shape)
            slice1 = [slice(None)] * nd
            slice2 = [slice(None)] * nd
            slice1[axis] = slice(1, None)
            slice2[axis] = slice(None, -1)
            slice1 = tuple(slice1)
            slice2 = tuple(slice2)
            if n > 1:
                return dask_diff(arr[slice1] - arr[slice2], n - 1, axis=axis)
            else:
                return arr[slice1] - arr[slice2]

        current_data = self._lazy_data(axis=axis, rechunk=rechunk)
        new_data = dask_diff(current_data, order, arr_axis)
        if not new_data.ndim:
            new_data = new_data.reshape((1, ))

        s = out or self._deepcopy_with_new_data(new_data)
        if out:
            if out.data.shape == new_data.shape:
                out.data = new_data
            else:
                raise ValueError(
                    "The output shape %s does not match  the shape of "
                    "`out` %s" % (new_data.shape, out.data.shape))
        axis2 = s.axes_manager[axis]
        new_offset = self.axes_manager[axis].offset + (order * axis2.scale / 2)
        axis2.offset = new_offset
        s.get_dimensions_from_data()
        if out is None:
            return s
        else:
            out.events.data_changed.trigger(obj=out)

    diff.__doc__ = BaseSignal.diff.__doc__

    def integrate_simpson(self, axis, out=None):
        axis = self.axes_manager[axis]
        from scipy import integrate
        axis = self.axes_manager[axis]
        data = self._lazy_data(axis=axis, rechunk=True)
        new_data = data.map_blocks(
            integrate.simps,
            x=axis.axis,
            axis=axis.index_in_array,
            drop_axis=axis.index_in_array,
            dtype=data.dtype)
        s = out or self._deepcopy_with_new_data(new_data)
        if out:
            if out.data.shape == new_data.shape:
                out.data = new_data
                out.events.data_changed.trigger(obj=out)
            else:
                raise ValueError(
                    "The output shape %s does not match  the shape of "
                    "`out` %s" % (new_data.shape, out.data.shape))
        else:
            s._remove_axis(axis.index_in_axes_manager)
            return s

    integrate_simpson.__doc__ = BaseSignal.integrate_simpson.__doc__

    def valuemax(self, axis, out=None, rechunk=True):
        idx = self.indexmax(axis, rechunk=rechunk)
        old_data = idx.data
        data = old_data.map_blocks(
            lambda x: self.axes_manager[axis].index2value(x))
        if out is None:
            idx.data = data
            return idx
        else:
            out.data = data
            out.events.data_changed.trigger(obj=out)

    valuemax.__doc__ = BaseSignal.valuemax.__doc__

    def valuemin(self, axis, out=None, rechunk=True):
        idx = self.indexmin(axis, rechunk=rechunk)
        old_data = idx.data
        data = old_data.map_blocks(
            lambda x: self.axes_manager[axis].index2value(x))
        if out is None:
            idx.data = data
            return idx
        else:
            out.data = data
            out.events.data_changed.trigger(obj=out)

    valuemin.__doc__ = BaseSignal.valuemin.__doc__

    def get_histogram(self, bins='fd', out=None, rechunk=True, **kwargs):
        if 'range_bins' in kwargs:
            _logger.warning("'range_bins' argument not supported for lazy "
                            "signals")
            del kwargs['range_bins']
        from hyperspy.signals import Signal1D
        data = self._lazy_data(rechunk=rechunk).flatten()
        hist, bin_edges = histogram_dask(data, bins=bins, **kwargs)
        if out is None:
            hist_spec = Signal1D(hist)
            hist_spec._lazy = True
            hist_spec._assign_subclass()
        else:
            hist_spec = out
            # we always overwrite the data because the computation is lazy ->
            # the result signal is lazy. Assume that the `out` is already lazy
            hist_spec.data = hist
        hist_spec.axes_manager[0].scale = bin_edges[1] - bin_edges[0]
        hist_spec.axes_manager[0].offset = bin_edges[0]
        hist_spec.axes_manager[0].size = hist.shape[-1]
        hist_spec.axes_manager[0].name = 'value'
        hist_spec.metadata.General.title = (
            self.metadata.General.title + " histogram")
        hist_spec.metadata.Signal.binned = True
        if out is None:
            return hist_spec
        else:
            out.events.data_changed.trigger(obj=out)

    get_histogram.__doc__ = BaseSignal.get_histogram.__doc__

    @staticmethod
    def _estimate_poissonian_noise_variance(dc, gain_factor, gain_offset,
                                            correlation_factor):
        variance = (dc * gain_factor + gain_offset) * correlation_factor
        # The lower bound of the variance is the gaussian noise.
        variance = da.clip(variance, gain_offset * correlation_factor, np.inf)
        return variance

    # def _get_navigation_signal(self, data=None, dtype=None):
    # return super()._get_navigation_signal(data=data, dtype=dtype).as_lazy()

    # _get_navigation_signal.__doc__ = BaseSignal._get_navigation_signal.__doc__

    # def _get_signal_signal(self, data=None, dtype=None):
    #     return super()._get_signal_signal(data=data, dtype=dtype).as_lazy()

    # _get_signal_signal.__doc__ = BaseSignal._get_signal_signal.__doc__

    def _calculate_summary_statistics(self, rechunk=True):
        if rechunk is True:
            # Use dask auto rechunk instead of HyperSpy's one, what should be
            # better for these operations
            rechunk = "dask_auto"
        data = self._lazy_data(rechunk=rechunk)
        _raveled = data.ravel()
        _mean, _std, _min, _q1, _q2, _q3, _max = da.compute(
            da.nanmean(data),
            da.nanstd(data),
            da.nanmin(data),
            da.percentile(_raveled, [25, ]),
            da.percentile(_raveled, [50, ]),
            da.percentile(_raveled, [75, ]),
            da.nanmax(data), )
        return _mean, _std, _min, _q1, _q2, _q3, _max

    def _map_all(self, function, inplace=True, **kwargs):
        calc_result = dd(function)(self.data, **kwargs)
        if inplace:
            self.data = da.from_delayed(calc_result, shape=self.data.shape,
                                        dtype=self.data.dtype)
            return None
        return self._deepcopy_with_new_data(calc_result)

    def _map_iterate(self,
                     function,
                     iterating_kwargs=(),
                     show_progressbar=None,
                     parallel=None,
                     max_workers=None,
                     ragged=None,
                     inplace=True,
                     **kwargs):
        if ragged not in (True, False):
            raise ValueError('"ragged" kwarg has to be bool for lazy signals')
        _logger.debug("Entering '_map_iterate'")

        size = max(1, self.axes_manager.navigation_size)
        from hyperspy.misc.utils import (create_map_objects,
                                         map_result_construction)
        func, iterators = create_map_objects(function, size, iterating_kwargs,
                                             **kwargs)
        iterators = (self._iterate_signal(), ) + iterators
        res_shape = self.axes_manager._navigation_shape_in_array
        # no navigation
        if not len(res_shape) and ragged:
            res_shape = (1,)

        all_delayed = [dd(func)(data) for data in zip(*iterators)]

        if ragged:
            if inplace:
                raise ValueError("In place computation is not compatible with "
                                  "ragged array for lazy signal.")
            # Shape of the signal dimension will change for the each nav. 
            # index, which means we can't predict the shape and the dtype needs
            # to be python object to support numpy ragged array
            sig_shape = ()
            sig_dtype = np.dtype('O')
        else:
            one_compute = all_delayed[0].compute()
            # No signal dimension for scalar
            if np.isscalar(one_compute):
                sig_shape = ()
                sig_dtype = type(one_compute)
            else:
                sig_shape = one_compute.shape
                sig_dtype = one_compute.dtype
        pixels = [
            da.from_delayed(
                res, shape=sig_shape, dtype=sig_dtype) for res in all_delayed
        ]
        if ragged:
            if show_progressbar is None:
                from hyperspy.defaults_parser import preferences
                show_progressbar = preferences.General.show_progressbar
            # We compute here because this is not sure if this is possible
            # to make a ragged dask array: we need to provide a chunk size...
            res_data = np.empty(res_shape, dtype=sig_dtype)
            _logger.info("Lazy signal is computed to make the ragged array.")
            if show_progressbar:
                cm = ProgressBar
            else:
                cm = dummy_context_manager
            with cm():
                try:
                    for i, pixel in enumerate(pixels):
                        res_data.flat[i] = pixel.compute()
                except MemoryError:
                    raise MemoryError("The use of 'ragged' array requires the "
                                      "computation of the lazy signal.")
        else:
            if len(pixels) > 0:
                for step in reversed(res_shape):
                    _len = len(pixels)
                    starts = range(0, _len, step)
                    ends = range(step, _len + step, step)
                    pixels = [
                        da.stack(
                            pixels[s:e], axis=0) for s, e in zip(starts, ends)
                    ]
            res_data = pixels[0]
    
        res = map_result_construction(
            self, inplace, res_data, ragged, sig_shape, lazy=not ragged)

        return res

    def _iterate_signal(self):
        if self.axes_manager.navigation_size < 2:
            yield self()
            return
        nav_dim = self.axes_manager.navigation_dimension
        sig_dim = self.axes_manager.signal_dimension
        nav_indices = self.axes_manager.navigation_indices_in_array[::-1]
        nav_lengths = np.atleast_1d(
            np.array(self.data.shape)[list(nav_indices)])
        getitem = [slice(None)] * (nav_dim + sig_dim)
        data = self._lazy_data()
        for indices in product(*[range(l) for l in nav_lengths]):
            for res, ind in zip(indices, nav_indices):
                getitem[ind] = res
            yield data[tuple(getitem)]

    def _block_iterator(self,
                        flat_signal=True,
                        get=threaded.get,
                        navigation_mask=None,
                        signal_mask=None):
        """A function that allows iterating lazy signal data by blocks,
        defining the dask.Array.

        Parameters
        ----------
        flat_signal: bool
            returns each block flattened, such that the shape (for the
            particular block) is (navigation_size, signal_size), with
            optionally masked elements missing. If false, returns
            the equivalent of s.inav[{blocks}].data, where masked elements are
            set to np.nan or 0.
        get : dask scheduler
            the dask scheduler to use for computations;
            default `dask.threaded.get`
        navigation_mask : {BaseSignal, numpy array, dask array}
            The navigation locations marked as True are not returned (flat) or
            set to NaN or 0.
        signal_mask : {BaseSignal, numpy array, dask array}
            The signal locations marked as True are not returned (flat) or set
            to NaN or 0.

        """
        self._make_lazy()
        data = self._data_aligned_with_axes
        nav_chunks = data.chunks[:self.axes_manager.navigation_dimension]
        indices = product(*[range(len(c)) for c in nav_chunks])
        signalsize = self.axes_manager.signal_size
        sig_reshape = (signalsize,) if signalsize else ()
        data = data.reshape((self.axes_manager.navigation_shape[::-1] +
                             sig_reshape))

        if signal_mask is None:
            signal_mask = slice(None) if flat_signal else \
                np.zeros(self.axes_manager.signal_size, dtype='bool')
        else:
            try:
                signal_mask = to_array(signal_mask).ravel()
            except ValueError:
                # re-raise with a message
                raise ValueError("signal_mask has to be a signal, numpy or"
                                 " dask array, but "
                                 "{} was given".format(type(signal_mask)))
            if flat_signal:
                signal_mask = ~signal_mask

        if navigation_mask is None:
            nav_mask = da.zeros(
                self.axes_manager.navigation_shape[::-1],
                chunks=nav_chunks,
                dtype='bool')
        else:
            try:
                nav_mask = to_array(navigation_mask, chunks=nav_chunks)
            except ValueError:
                # re-raise with a message
                raise ValueError("navigation_mask has to be a signal, numpy or"
                                 " dask array, but "
                                 "{} was given".format(type(navigation_mask)))
        if flat_signal:
            nav_mask = ~nav_mask
        for ind in indices:
            chunk = get(data.dask,
                        (data.name, ) + ind + (0,) * bool(signalsize))
            n_mask = get(nav_mask.dask, (nav_mask.name, ) + ind)
            if flat_signal:
                yield chunk[n_mask, ...][..., signal_mask]
            else:
                chunk = chunk.copy()
                value = np.nan if np.can_cast('float', chunk.dtype) else 0
                chunk[n_mask, ...] = value
                chunk[..., signal_mask] = value
                yield chunk.reshape(chunk.shape[:-1] +
                                    self.axes_manager.signal_shape[::-1])

    def decomposition(
        self,
        normalize_poissonian_noise=False,
        algorithm="SVD",
        output_dimension=None,
        signal_mask=None,
        navigation_mask=None,
        get=threaded.get,
        num_chunks=None,
        reproject=True,
        print_info=True,
        **kwargs
    ):
        """Perform Incremental (Batch) decomposition on the data.

        The results are stored in ``self.learning_results``.

        Read more in the :ref:`User Guide <big_data.decomposition>`.

        Parameters
        ----------
        normalize_poissonian_noise : bool, default False
            If True, scale the signal to normalize Poissonian noise using
            the approach described in [KeenanKotula2004]_.
        algorithm : {'SVD', 'PCA', 'ORPCA', 'ORNMF'}, default 'SVD'
            The decomposition algorithm to use.
        output_dimension : int or None, default None
            Number of components to keep/calculate. If None, keep all
            (only valid for 'SVD' algorithm)
        get : dask scheduler
            the dask scheduler to use for computations;
            default `dask.threaded.get`
        num_chunks : int or None, default None
            the number of dask chunks to pass to the decomposition model.
            More chunks require more memory, but should run faster. Will be
            increased to contain at least ``output_dimension`` signals.
        navigation_mask : {BaseSignal, numpy array, dask array}
            The navigation locations marked as True are not used in the
            decomposition.
        signal_mask : {BaseSignal, numpy array, dask array}
            The signal locations marked as True are not used in the
            decomposition.
        reproject : bool, default True
            Reproject data on the learnt components (factors) after learning.
        print_info : bool, default True
            If True, print information about the decomposition being performed.
            In the case of sklearn.decomposition objects, this includes the
            values of all arguments of the chosen sklearn algorithm.
        **kwargs
            passed to the partial_fit/fit functions.

        References
        ----------
        .. [KeenanKotula2004] M. Keenan and P. Kotula, "Accounting for Poisson noise
            in the multivariate analysis of ToF-SIMS spectrum images", Surf.
            Interface Anal 36(3) (2004): 203-212.

        See Also
        --------
        * :py:meth:`~.learn.mva.MVA.decomposition` for non-lazy signals
        * :py:func:`dask.array.linalg.svd`
        * :py:class:`sklearn.decomposition.IncrementalPCA`
        * :py:class:`~.learn.rpca.ORPCA`
        * :py:class:`~.learn.ornmf.ORNMF`

        """
        if kwargs.get("bounds", False):
            warnings.warn(
                "The `bounds` keyword is deprecated and will be removed "
                "in v2.0. Since version > 1.3 this has no effect.",
                VisibleDeprecationWarning,
            )
            kwargs.pop("bounds", None)

        # Deprecate 'ONMF' for 'ORNMF'
        if algorithm == "ONMF":
            warnings.warn(
                "The argument `algorithm='ONMF'` has been deprecated and will "
                "be removed in future. Please use `algorithm='ORNMF'` instead.",
                VisibleDeprecationWarning,
            )
            algorithm = "ORNMF"


        # Check algorithms requiring output_dimension
        algorithms_require_dimension = ["PCA", "ORPCA", "ORNMF"]
        if algorithm in algorithms_require_dimension and output_dimension is None:
            raise ValueError(
                "`output_dimension` must be specified for '{}'".format(algorithm)
            )

        explained_variance = None
        explained_variance_ratio = None

        _al_data = self._data_aligned_with_axes
        nav_chunks = _al_data.chunks[: self.axes_manager.navigation_dimension]
        sig_chunks = _al_data.chunks[self.axes_manager.navigation_dimension :]

        num_chunks = 1 if num_chunks is None else num_chunks
        blocksize = np.min([multiply(ar) for ar in product(*nav_chunks)])
        nblocks = multiply([len(c) for c in nav_chunks])

        if output_dimension and blocksize / output_dimension < num_chunks:
            num_chunks = np.ceil(blocksize / output_dimension)

        blocksize *= num_chunks

        # Initialize return_info and print_info
        to_return = None
        to_print = [
            "Decomposition info:",
            "  normalize_poissonian_noise={}".format(normalize_poissonian_noise),
            "  algorithm={}".format(algorithm),
            "  output_dimension={}".format(output_dimension)
        ]

        # LEARN
        if algorithm == "PCA":
            if not import_sklearn.sklearn_installed:
                raise ImportError("algorithm='PCA' requires scikit-learn")

            obj = import_sklearn.sklearn.decomposition.IncrementalPCA(n_components=output_dimension)
            method = partial(obj.partial_fit, **kwargs)
            reproject = True
            to_print.extend(["scikit-learn estimator:", obj])

        elif algorithm == "ORPCA":
            from hyperspy.learn.rpca import ORPCA

            batch_size = kwargs.pop("batch_size", None)
            obj = ORPCA(output_dimension, **kwargs)
            method = partial(obj.fit, batch_size=batch_size)

        elif algorithm == "ORNMF":
            from hyperspy.learn.ornmf import ORNMF

            batch_size = kwargs.pop("batch_size", None)
            obj = ORNMF(output_dimension, **kwargs)
            method = partial(obj.fit, batch_size=batch_size)

        elif algorithm != "SVD":
            raise ValueError("'algorithm' not recognised")

        original_data = self.data
        try:
            _logger.info("Performing decomposition analysis")

            if normalize_poissonian_noise:
                _logger.info("Scaling the data to normalize Poissonian noise")

                data = self._data_aligned_with_axes
                ndim = self.axes_manager.navigation_dimension
                sdim = self.axes_manager.signal_dimension
                nm = da.logical_not(
                    da.zeros(self.axes_manager.navigation_shape[::-1], chunks=nav_chunks)
                    if navigation_mask is None
                    else to_array(navigation_mask, chunks=nav_chunks)
                )
                sm = da.logical_not(
                    da.zeros(self.axes_manager.signal_shape[::-1], chunks=sig_chunks)
                    if signal_mask is None
                    else to_array(signal_mask, chunks=sig_chunks)
                )
                ndim = self.axes_manager.navigation_dimension
                sdim = self.axes_manager.signal_dimension
                bH, aG = da.compute(
                    data.sum(axis=tuple(range(ndim))),
                    data.sum(axis=tuple(range(ndim, ndim + sdim))),
                )
                bH = da.where(sm, bH, 1)
                aG = da.where(nm, aG, 1)

                raG = da.sqrt(aG)
                rbH = da.sqrt(bH)

                coeff = raG[(...,) + (None,) * rbH.ndim] * rbH[(None,) * raG.ndim + (...,)]
                coeff.map_blocks(np.nan_to_num)
                coeff = da.where(coeff == 0, 1, coeff)
                data = data / coeff
                self.data = data

            # LEARN
            if algorithm == "SVD":
                reproject = False
                from dask.array.linalg import svd

                try:
                    self._unfolded4decomposition = self.unfold()
                    # TODO: implement masking
                    if navigation_mask or signal_mask:
                        raise NotImplementedError("Masking is not yet implemented for lazy SVD")

                    U, S, V = svd(self.data)

                    if output_dimension is None:
                        min_shape = min(min(U.shape), min(V.shape))
                    else:
                        min_shape = output_dimension

                    U = U[:, :min_shape]
                    S = S[:min_shape]
                    V = V[:min_shape]

                    factors = V.T
                    explained_variance = S ** 2 / self.data.shape[0]
                    loadings = U * S
                finally:
                    if self._unfolded4decomposition is True:
                        self.fold()
                        self._unfolded4decomposition is False
            else:
                this_data = []
                try:
                    for chunk in progressbar(
                        self._block_iterator(
                            flat_signal=True,
                            get=get,
                            signal_mask=signal_mask,
                            navigation_mask=navigation_mask,
                        ),
                        total=nblocks,
                        leave=True,
                        desc="Learn",
                    ):
                        this_data.append(chunk)
                        if len(this_data) == num_chunks:
                            thedata = np.concatenate(this_data, axis=0)
                            method(thedata)
                            this_data = []
                    if len(this_data):
                        thedata = np.concatenate(this_data, axis=0)
                        method(thedata)
                except KeyboardInterrupt:  # pragma: no cover
                    pass

            # GET ALREADY CALCULATED RESULTS
            if algorithm == "PCA":
                explained_variance = obj.explained_variance_
                explained_variance_ratio = obj.explained_variance_ratio_
                factors = obj.components_.T

            elif algorithm == "ORPCA":
                factors, loadings = obj.finish()
                loadings = loadings.T

            elif algorithm == "ORNMF":
                factors, loadings = obj.finish()
                loadings = loadings.T

            # REPROJECT
            if reproject:
                if algorithm == "PCA":
                    method = obj.transform

                    def post(a):
                        return np.concatenate(a, axis=0)

                elif algorithm == "ORPCA":
                    method = obj.project

                    def post(a):
                        return np.concatenate(a, axis=1).T

                elif algorithm == "ORNMF":
                    method = obj.project

                    def post(a):
                        return np.concatenate(a, axis=1).T

                _map = map(
                    lambda thing: method(thing),
                    self._block_iterator(
                        flat_signal=True,
                        get=get,
                        signal_mask=signal_mask,
                        navigation_mask=navigation_mask,
                    ),
                )
                H = []
                try:
                    for thing in progressbar(_map, total=nblocks, desc="Project"):
                        H.append(thing)
                except KeyboardInterrupt:  # pragma: no cover
                    pass
                loadings = post(H)

            if explained_variance is not None and explained_variance_ratio is None:
                explained_variance_ratio = explained_variance / explained_variance.sum()

            # RESHUFFLE "blocked" LOADINGS
            ndim = self.axes_manager.navigation_dimension
            if algorithm != "SVD":  # Only needed for online algorithms
                try:
                    loadings = _reshuffle_mixed_blocks(
                        loadings, ndim, (output_dimension,), nav_chunks
                    ).reshape((-1, output_dimension))
                except ValueError:
                    # In case the projection step was not finished, it's left
                    # as scrambled
                    pass
        finally:
            self.data = original_data

        target = self.learning_results
        target.decomposition_algorithm = algorithm
        target.output_dimension = output_dimension
        if algorithm != "SVD":
            target._object = obj
        target.factors = factors
        target.loadings = loadings
        target.explained_variance = explained_variance
        target.explained_variance_ratio = explained_variance_ratio

        # Rescale the results if the noise was normalized
        if normalize_poissonian_noise is True:
            target.factors = target.factors * rbH.ravel()[:, np.newaxis]
            target.loadings = target.loadings * raG.ravel()[:, np.newaxis]

        # Print details about the decomposition we just performed
        if print_info:
            print("\n".join([str(pr) for pr in to_print]))


def _reshuffle_mixed_blocks(array, ndim, sshape, nav_chunks):
    """Reshuffles dask block-shuffled array

    Parameters
    ----------
    array : np.ndarray
        the array to reshuffle
    ndim : int
        the number of navigation (shuffled) dimensions
    sshape : tuple of ints
        The shape
    """
    splits = np.cumsum([multiply(ar)
                        for ar in product(*nav_chunks)][:-1]).tolist()
    if splits:
        all_chunks = [
            ar.reshape(shape + sshape)
            for shape, ar in zip(
                product(*nav_chunks), np.split(array, splits))
        ]

        def split_stack_list(what, step, axis):
            total = len(what)
            if total != step:
                return [
                    np.concatenate(
                        what[i:i + step], axis=axis)
                    for i in range(0, total, step)
                ]
            else:
                return np.concatenate(what, axis=axis)

        for chunks, axis in zip(nav_chunks[::-1], range(ndim - 1, -1, -1)):
            step = len(chunks)
            all_chunks = split_stack_list(all_chunks, step, axis)
        return all_chunks
    else:
        return array



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy._components.expression import Expression


class Logistic(Expression):

    r"""Logistic function (sigmoid or s-shaped curve) component.

    .. math::
    
        f(x) = \frac{a}{1 + b\cdot \mathrm{exp}\left[-c 
            \left((x - x_0\right)\right]}

    ============== =============
    Variable        Parameter 
    ============== =============
    :math:`A`       a 
    :math:`b`       b 
    :math:`c`       c 
    :math:`x_0`     origin 
    ============== =============


    Parameters
    -----------
    a : Float
        The curve's maximum y-value,  
        :math:`\mathrm{lim}_{x\to\infty}\left(y\right) = a`
    b : Float
        Additional parameter: 
        b>1 shifts origin to larger values;
        0<b<1 shifts origin to smaller values;
        b<0 introduces an asymptote
    c : Float
        Logistic growth rate or steepness of the curve
    origin : Float
        Position of the sigmoid's midpoint
    **kwargs
        Extra keyword arguments are passed to the ``Expression`` component.
    """

    def __init__(self, a=1., b=1., c=1., origin=0., module="numexpr", **kwargs):
        super(Logistic, self).__init__(
            expression="a / (1 + b * exp(-c * (x - origin)))",
            name="Logistic",
            a=a,
            b=b,
            c=c,
            origin=origin,
            position="origin",
            module=module,
            autodoc=False,
            **kwargs)

        # Boundaries
        self.isbackground = False



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np

from hyperspy.components1d import Gaussian
from hyperspy.misc.utils import DictionaryTreeBrowser
from hyperspy.samfire_utils.strategy import GlobalStrategy, LocalStrategy
from hyperspy.signals import Signal1D


class someweight(object):

    def __init__(self):
        self.model = None

    def map(self, calc_pixels, slices=slice(None, None)):
        # return of slices shape thing, with np.nan where mask is False
        thing = self.model.chisq.data[slices].copy()
        thing[np.logical_not(calc_pixels)] = np.nan
        return thing

    def function(self, ind):
        return self.model.chisq.data[ind]


def create_artificial_samfire(shape):
    artificial_samfire = DictionaryTreeBrowser()
    artificial_samfire.add_node('running_pixels')
    artificial_samfire.running_pixels = []
    artificial_samfire.add_node('model')
    artificial_samfire.add_node('metadata')
    artificial_samfire.metadata.add_node('marker')
    artificial_samfire.metadata.marker = np.zeros(shape)
    artificial_samfire.add_node('_scale')
    artificial_samfire._scale = 1.0
    return artificial_samfire


def compare_two_value_dicts(ans_r, ans):
    test = True
    for k, v in ans_r.items():
        test = test and k in ans
        if test:
            for p, pv in v.items():
                test = test and p in ans[k]
                if test:
                    np.testing.assert_allclose(
                        np.array(pv),
                        np.array(
                            ans[k][p]))
    return test


class TestLocalSimple:

    def setup_method(self, method):
        self.shape = (5, 7)
        self.s = LocalStrategy('test diffusion strategy')
        self.samf = create_artificial_samfire(self.shape)

        m = DictionaryTreeBrowser()
        m.set_item('chisq.data', np.ones(self.shape) * 5.)

        self.samf.model = m

    def test_default_init(self):
        s = self.s
        assert s.samf is None
        assert s.radii is None
        assert s.weight is None
        assert s._untruncated is None
        assert s._mask_all is None
        assert s._radii_changed

    def test_samf_weight_setters(self):
        s = self.s
        samf = self.samf
        w = someweight()
        s.weight = w
        assert s._weight is w
        assert w.model is None
        s.samf = samf
        assert w.model is samf.model
        w2 = someweight()
        s.weight = w2
        assert w.model is None

    def test_radii(self):
        s = self.s
        s.samf = self.samf
        assert s.radii is None
        assert s._radii_changed
        s._radii_changed = False
        s.radii = 1.
        assert s.radii == (1.0, 1.0)
        assert s._radii_changed

    def test_clean(self):
        s = self.s
        s._untruncated = 12
        s._mask_all = 2.
        s._radii_changed = False
        s.clean()
        assert s._untruncated is None
        assert s._mask_all is None
        assert s._radii_changed

    def test_refresh_overwrite(self):
        s = self.s
        s.radii = (1.9, 2.1)
        samf = self.samf
        s.samf = samf
        s.decay_function = lambda x: np.exp(-x)
        s.weight = someweight()
        s.samf.metadata.marker[1, 1] = -2
        s.samf.metadata.marker[0, 1] = -1
        s.samf.model.chisq.data[0, 1] = 50

        ans1 = np.array([[1.63810767e-03, -1.00000000e+00, 1.63810767e-03, 2.61027907e-23],
                         [2.47875218e-03, -
                          1.00000000e+00, 2.47875218e-03, 9.11881966e-04],
                         [1.63810767e-03,
                          2.47875218e-03,
                          1.63810767e-03,
                          0.00000000e+00],
                         [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])

        s.refresh(True, given_pixels=None)
        np.testing.assert_allclose(ans1, s.samf.metadata.marker[:4, :4])

        given = np.ones(self.shape, dtype=bool)
        given[0, 1] = False
        s.refresh(True, given_pixels=given)
        np.testing.assert_allclose(
            s.samf.metadata.marker[
                ~given][0],
            0.011624353837970535)
        assert np.all(s.samf.metadata.marker[given] == -1)

    def test_refresh_nooverwrite(self):
        s = self.s
        s.radii = (1.9, 2.1)
        samf = self.samf
        s.samf = samf
        s.decay_function = lambda x: np.exp(-x)
        s.weight = someweight()
        s.samf.metadata.marker[1, 1] = -2
        s.samf.metadata.marker[0, 1] = -1
        s.samf.model.chisq.data[0, 1] = 50

        ans1 = np.array([[7.09547416e-23, -1.00000000e+00, 7.09547416e-23, 2.61027907e-23],
                         [4.68911365e-23, -
                          2.00000000e+00, 4.68911365e-23, 0.00000000e+00],
                         [0.00000000e+00,
                          0.00000000e+00,
                          0.00000000e+00,
                          0.00000000e+00],
                         [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])

        s.refresh(False, given_pixels=None)
        np.testing.assert_allclose(ans1, s.samf.metadata.marker[:4, :4])

        s.samf.metadata.marker[0, 0] = -1
        given = np.ones(self.shape, dtype=bool)
        given[0, 0] = False
        # should stay the same, as the new point [0,0] is not in "given"
        s.refresh(False, given_pixels=given)
        np.testing.assert_allclose(ans1, s.samf.metadata.marker[:4, :4])

    def test_get_distance_array(self):
        s = self.s
        s.samf = self.samf
        # use floats intentionally, since it was broken at one point
        s.radii = (2.1, 1.9)
        ind = (0, 0)
        distances, slices, centre, mask = s._get_distance_array(
            self.shape, ind)
        assert centre == ind
        assert slices == (slice(0, 4, None), slice(0, 3, None))
        tmp = np.array([[0., 1., np.nan],
                        [1., 1.41421356, np.nan],
                        [2., np.nan, np.nan],
                        [np.nan, np.nan, np.nan]])
        tmp_m = np.array([[False, True, False],
                          [True, True, False],
                          [True, False, False],
                          [False, False, False]], dtype=bool)
        assert np.all(tmp_m == mask)
        np.testing.assert_allclose(tmp[mask], distances[mask])
        assert not s._radii_changed
        tmp_ma = np.array([[3.14884957, 2.31782464, 2.04081633, 2.31782464, 3.14884957],
                           [2.01506272,
                            1.18403779,
                            0.90702948,
                            1.18403779,
                            2.01506272],
                           [1.33479061,
                            0.50376568,
                            0.22675737,
                            0.50376568,
                            1.33479061],
                           [1.10803324,
                            0.27700831,
                            0.,
                            0.27700831,
                            1.10803324],
                           [1.33479061,
                            0.50376568,
                            0.22675737,
                            0.50376568,
                            1.33479061],
                           [2.01506272,
                            1.18403779,
                            0.90702948,
                            1.18403779,
                            2.01506272],
                           [3.14884957, 2.31782464, 2.04081633, 2.31782464, 3.14884957]])
        np.testing.assert_allclose(s._mask_all, tmp_ma)
        tmp_un = np.array([[3.60555128, 3.16227766, 3., 3.16227766, 3.60555128],
                           [2.82842712,
                            2.23606798,
                            2.,
                            2.23606798,
                            2.82842712],
                           [2.23606798,
                            1.41421356,
                            1.,
                            1.41421356,
                            2.23606798],
                           [2., 1., 0., 1., 2.],
                           [2.23606798,
                            1.41421356,
                            1.,
                            1.41421356,
                            2.23606798],
                           [2.82842712,
                            2.23606798,
                            2.,
                            2.23606798,
                            2.82842712],
                           [3.60555128, 3.16227766, 3., 3.16227766, 3.60555128]])
        np.testing.assert_allclose(s._untruncated, tmp_un)

        # now check that the stored values are used
        # mask:
        s._mask_all[0, 0] = 1.0
        ind = (4, 6)
        distances, slices, centre, mask = s._get_distance_array(
            self.shape, ind)
        assert centre == (3.0, 2.0)
        assert slices == (slice(1, 5, None), slice(4, 7, None))
        tmp_m = np.array([[True, False, False],
                          [False, False, True],
                          [False, True, True],
                          [False, True, False]], dtype=bool)
        tmp = np.array([[3.60555128, np.nan, np.nan],
                        [np.nan, np.nan, 2.],
                        [np.nan, 1.41421356, 1.],
                        [np.nan, 1., 0.]])
        assert np.all(tmp_m == mask)
        np.testing.assert_allclose(tmp[mask], distances[mask])
        assert not s._radii_changed

        # now mask radii changed and check that the correct result is
        # calculated again
        s._radii_changed = True

        distances, slices, centre, mask = s._get_distance_array(
            self.shape, ind)
        assert centre == (3.0, 2.0)
        assert slices == (slice(1, 5, None), slice(4, 7, None))
        tmp_m = np.array([[False, False, False],
                          [False, False, True],
                          [False, True, True],
                          [False, True, False]], dtype=bool)
        tmp = np.array([[np.nan, np.nan, np.nan],
                        [np.nan, np.nan, 2.],
                        [np.nan, 1.41421356, 1.],
                        [np.nan, 1., 0.]])
        assert np.all(tmp_m == mask)
        np.testing.assert_allclose(tmp[mask], distances[mask])
        assert not s._radii_changed

    def test_update_marker(self):
        s = self.s
        # use floats intentionally, since it was broken at one point
        s.radii = (1.9, 2.1)
        samf = self.samf
        samf.metadata.marker[0, 1] = -0.33
        samf.metadata.marker[1, 2] = 100
        s.samf = samf
        s.decay_function = lambda x: np.exp(-x)
        s.weight = someweight()

        ind = (0, 0)
        s._update_marker(ind)
        tmp_m1 = np.array([[-1.00000000e+00, -3.30000000e-01, 9.11881966e-04, 0.00000000e+00],
                           [2.47875218e-03,
                            1.63810767e-03,
                            1.00000000e+02,
                            0.00000000e+00],
                           [0.00000000e+00,
                            0.00000000e+00,
                            0.00000000e+00,
                            0.00000000e+00],
                           [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])
        np.testing.assert_allclose(tmp_m1, s.samf.metadata.marker[:4, :4])

        ind = (1, 1)
        s.samf.running_pixels.append((1, 2))
        s.samf._scale = 13
        s._update_marker(ind)
        tmp_m2 = np.array([[-1.00000000e+00, -3.30000000e-01, 2.54998964e-03, 0.00000000e+00],
                           [4.95750435e-03, -
                            1.30000000e+01, 0.00000000e+00, 9.11881966e-04],
                           [1.63810767e-03,
                            2.47875218e-03,
                            1.63810767e-03,
                            0.00000000e+00],
                           [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])
        np.testing.assert_allclose(tmp_m2, s.samf.metadata.marker[:4, :4])


class TestLocalWithModel:

    def setup_method(self, method):
        self.shape = (5, 7)
        self.s = LocalStrategy('test diffusion strategy')
        self.samf = create_artificial_samfire(self.shape)

        m = Signal1D(np.empty(self.shape + (100,))).create_model()
        m.extend([Gaussian() for _ in range(3)])
        m.chisq.data.fill(5.)

        self.samf.model = m

    def test_values(self):
        s = self.s
        # use floats intentionally, since it was broken at one point
        s.radii = (1.9, 2.1)
        samf = self.samf
        s.samf = samf
        s.decay_function = lambda x: np.exp(-x)
        s.weight = someweight()

        samf.model[0].active = False

        samf.model[1].active_is_multidimensional = True
        samf.model[1]._active_array[0, 0] = False
        samf.model[1].A.map['values'].fill(10.)
        samf.model[1].centre.map['values'].fill(3.)
        samf.model[1].centre.map['values'][0, 1] = 15.

        samf.model[2].A.free = False

        d1 = s.values((0, 0))
        assert d1 == {}

        samf.metadata.marker[0, 0] = -1
        assert (
            s.values(
                (1, 0)) == {
                'Gaussian_1': {
                    'centre': 0.0, 'sigma': 0.0}})

        samf.metadata.marker[1, 1] = -1

        ans_r = {'Gaussian_0': {'A': 10.0, 'centre': 2.9999999999999996, 'sigma': 0.0},
                 'Gaussian_1': {'centre': 0.0, 'sigma': 0.0}}

        ans = s.values((1, 0))

        test = compare_two_value_dicts(ans_r, ans)
        assert test

        samf.metadata.marker[0, 1] = -1

        ans_r2 = {'Gaussian_0': {'A': 10.0, 'centre': 7.7748266350745405, 'sigma': 0.0},
                  'Gaussian_1': {'centre': 0.0, 'sigma': 0.0}}

        ans2 = s.values((1, 0))

        test2 = compare_two_value_dicts(ans_r2, ans2)
        assert test2


class TestGlobalStrategy:

    def setup_method(self, method):
        # TODO: actually finish setup+ tests
        self.shape = (5, 7)
        self.s = GlobalStrategy('test segmenter strategy')
        self.samf = create_artificial_samfire(self.shape)

        m = Signal1D(np.empty(self.shape + (100,))).create_model()
        m.extend([Gaussian() for _ in range(3)])
        m.chisq.data.fill(5.)

        self.samf.model = m

    def test_refresh_nooverwrite_nogiven(self):
        samf = self.samf
        samf.set_item('update_every', np.nan)
        s = self.s
        s.samf = samf

        samf.metadata.marker[0, 0] = -1
        samf.metadata.marker[0, 1] = -2

        s.refresh(False)
        assert samf.metadata.marker[0, 0] == -1
        assert samf.metadata.marker[0, 1] == -2
        assert np.all(samf.metadata.marker.ravel()[2:] == 1)

    def test_refresh_nooverwrite_given(self):
        samf = self.samf
        samf.set_item('update_every', np.nan)
        s = self.s
        s.samf = samf

        samf.metadata.marker.fill(3.)
        samf.metadata.marker[0, 0] = -1
        samf.metadata.marker[0, 1] = -2

        given = np.zeros(self.shape, dtype=bool)
        given[0, 1] = True
        s.refresh(False, given_pixels=given)
        assert samf.metadata.marker[0, 0] == 1
        assert samf.metadata.marker[0, 1] == -2
        assert np.all(samf.metadata.marker.ravel()[2:] == 1)

    def test_refresh_overwrite_nogiven(self):
        samf = self.samf
        samf.set_item('update_every', np.nan)
        s = self.s
        s.samf = samf

        samf.metadata.marker[0, 0] = -1
        samf.metadata.marker[0, 1] = -2

        s.refresh(True)
        assert samf.metadata.marker[0, 0] == -1
        assert samf.metadata.marker[0, 1] == -1
        assert np.all(samf.metadata.marker.ravel()[2:] == 1)

    def test_refresh_overwrite_given(self):
        samf = self.samf
        samf.set_item('update_every', np.nan)
        s = self.s
        s.samf = samf

        samf.metadata.marker.fill(3.)
        samf.metadata.marker[0, 0] = -1
        samf.metadata.marker[0, 1] = -2

        given = np.zeros(self.shape, dtype=bool)
        given[0, 1] = True
        s.refresh(True, given_pixels=given)
        assert samf.metadata.marker[0, 0] == 1
        assert samf.metadata.marker[0, 1] == -1
        assert np.all(samf.metadata.marker.ravel()[2:] == 1)

    def test_update_marker(self):
        s = self.s
        samf = self.samf
        s.samf = samf
        ind = (1, 1)
        s._update_marker(ind)
        assert samf.metadata.marker[ind] == -samf._scale
        mask = np.ones(self.shape, dtype=bool)
        mask[ind] = False
        assert np.all(samf.metadata.marker[mask] == 0)

    def test_package_values(self):

        s = self.s
        samf = self.samf
        s.samf = samf

        samf.model[0].active = False
        samf.model[1].active_is_multidimensional = True
        samf.model[1]._active_array[0, 0] = False
        samf.model[1].A.map['values'].fill(10.)
        samf.model[1].centre.map['values'].fill(3.)
        samf.model[1].centre.map['values'][0, 1] = 15.
        samf.model[2].sigma.map['values'][0, 2] = 9.

        s.samf.metadata.marker[0, 0] = -100

        ans_r1 = {'Gaussian_0': {'A': np.array([], dtype=float),
                                 'centre': np.array([], dtype=float),
                                 'sigma': np.array([], dtype=float)},
                  'Gaussian_1': {'centre': np.array([0.]), 'sigma': np.array([0.])}}

        ans = s._package_values()
        t = compare_two_value_dicts(ans_r1, ans)
        assert t

        s.samf.metadata.marker[0, 2] = -100

        ans_r2 = {'Gaussian_0': {'A': np.array([10.]),
                                 'centre': np.array([3.]),
                                 'sigma': np.array([0.])},
                  'Gaussian_1': {'centre': np.array([0., 0.]), 'sigma': np.array([0., 9.])}}

        ans2 = s._package_values()
        t2 = compare_two_value_dicts(ans_r2, ans2)
        assert t2



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import gc
import os
import tempfile
import warnings

import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.io_plugins.blockfile import get_default_header
from hyperspy.misc.array_tools import sarray2dict
from hyperspy.misc.date_time_tools import serial_date_to_ISO_format
from hyperspy.misc.test_utils import assert_deep_almost_equal

try:
    WindowsError
except NameError:
    WindowsError = None


DIRPATH = os.path.dirname(__file__)
FILE1 = os.path.join(DIRPATH, 'blockfile_data', 'test1.blo')
FILE2 = os.path.join(DIRPATH, 'blockfile_data', 'test2.blo')


@pytest.fixture()
def save_path():
    with tempfile.TemporaryDirectory() as tmp:
        filepath = os.path.join(tmp, 'blockfile_data', 'save_temp.blo')
        yield filepath
        # Force files release (required in Windows)
        gc.collect()


ref_data2 = np.array(
    [[[[20, 23, 25, 25, 27],
       [29, 23, 23, 0, 29],
       [24, 0, 0, 22, 18],
       [0, 14, 19, 17, 26],
       [19, 21, 22, 27, 20]],

      [[28, 25, 29, 15, 29],
       [12, 15, 12, 25, 24],
       [25, 26, 26, 18, 27],
       [19, 18, 20, 23, 28],
       [28, 18, 22, 25, 0]],

      [[21, 29, 25, 19, 18],
       [30, 15, 20, 22, 26],
       [23, 18, 26, 15, 25],
       [22, 25, 24, 15, 20],
       [22, 15, 15, 21, 23]]],


     [[[28, 25, 26, 24, 26],
       [26, 17, 0, 24, 12],
       [17, 18, 21, 19, 21],
       [21, 24, 19, 17, 0],
       [17, 14, 25, 15, 26]],

      [[25, 18, 20, 15, 24],
       [19, 13, 23, 18, 11],
       [0, 25, 0, 0, 14],
       [26, 22, 22, 11, 14],
       [21, 0, 15, 13, 19]],

      [[24, 18, 20, 22, 21],
       [13, 25, 20, 28, 29],
       [15, 17, 24, 23, 23],
       [22, 21, 21, 22, 18],
       [24, 25, 18, 18, 27]]]], dtype=np.uint8)

axes1 = {
    'axis-0': {
        'name': 'y', 'navigate': True, 'offset': 0.0,
        'scale': 12.8, 'size': 3, 'units': 'nm'},
    'axis-1': {
        'name': 'x', 'navigate': True, 'offset': 0.0,
        'scale': 12.8, 'size': 2, 'units': 'nm'},
    'axis-2': {
        'name': 'dy', 'navigate': False, 'offset': 0.0,
        'scale': 0.016061676839061997, 'size': 144, 'units': 'cm'},
    'axis-3': {
        'name': 'dx', 'navigate': False, 'offset': 0.0,
        'scale': 0.016061676839061997, 'size': 144, 'units': 'cm'}}

axes2 = {
    'axis-0': {
        'name': 'y', 'navigate': True, 'offset': 0.0,
        'scale': 64.0, 'size': 2, 'units': 'nm'},
    'axis-1': {
        'name': 'x', 'navigate': True, 'offset': 0.0,
        'scale': 64.0, 'size': 3, 'units': 'nm'},
    'axis-2': {
        'name': 'dy', 'navigate': False, 'offset': 0.0,
        'scale': 0.016061676839061997, 'size': 5, 'units': 'cm'},
    'axis-3': {
        'name': 'dx', 'navigate': False, 'offset': 0.0,
        'scale': 0.016061676839061997, 'size': 5, 'units': 'cm'}}

axes2_converted = {
    'axis-0': {
        'name': 'y', 'navigate': True, 'offset': 0.0,
        'scale': 64.0, 'size': 2, 'units': 'nm'},
    'axis-1': {
        'name': 'x', 'navigate': True, 'offset': 0.0,
        'scale': 64.0, 'size': 3, 'units': 'nm'},
    'axis-2': {
        'name': 'dy', 'navigate': False, 'offset': 0.0,
        'scale': 160.61676839061997, 'size': 5, 'units': 'µm'},
    'axis-3': {
        'name': 'dx', 'navigate': False, 'offset': 0.0,
        'scale': 160.61676839061997, 'size': 5, 'units': 'µm'}}


def test_load1():
    s = hs.load(FILE1)
    assert s.data.shape == (3, 2, 144, 144)
    assert s.axes_manager.as_dictionary() == axes1


@pytest.mark.parametrize(("convert_units"), (True, False))
def test_load2(convert_units):
    s = hs.load(FILE2, convert_units=convert_units)
    assert s.data.shape == (2, 3, 5, 5)
    axes = axes2_converted if convert_units else axes2
    np.testing.assert_equal(s.axes_manager.as_dictionary(), axes)
    np.testing.assert_allclose(s.data, ref_data2)


@pytest.mark.parametrize(("convert_units"), (True, False))
def test_save_load_cycle(save_path, convert_units):
    sig_reload = None
    signal = hs.load(FILE2, convert_units=convert_units)
    serial = signal.original_metadata['blockfile_header']['Acquisition_time']
    date, time, timezone = serial_date_to_ISO_format(serial)
    assert signal.metadata.General.original_filename == 'test2.blo'
    assert signal.metadata.General.date == date
    assert signal.metadata.General.time == time
    assert signal.metadata.General.time_zone == timezone
    assert (
        signal.metadata.General.notes ==
        "Precession angle : \r\nPrecession Frequency : \r\nCamera gamma : on")
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path, convert_units=convert_units)
    np.testing.assert_equal(signal.data, sig_reload.data)
    assert (signal.axes_manager.as_dictionary() ==
            sig_reload.axes_manager.as_dictionary())
    assert (signal.original_metadata.as_dictionary() ==
            sig_reload.original_metadata.as_dictionary())
    # change original_filename to make the metadata of both signals equals
    sig_reload.metadata.General.original_filename = signal.metadata.General.original_filename
    assert_deep_almost_equal(signal.metadata.as_dictionary(),
                             sig_reload.metadata.as_dictionary())
    assert (
        signal.metadata.General.date ==
        sig_reload.metadata.General.date)
    assert (
        signal.metadata.General.time ==
        sig_reload.metadata.General.time)
    assert isinstance(signal, hs.signals.Signal2D)
    # Delete reference to close memmap file!
    del sig_reload


def test_different_x_y_scale_units(save_path):
    # perform load and save cycle with changing the scale on y
    signal = hs.load(FILE2)
    signal.axes_manager[0].scale = 50.0
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(sig_reload.axes_manager[0].scale, 50.0,
                    rtol=1E-5)
    np.testing.assert_allclose(sig_reload.axes_manager[1].scale, 64.0,
                    rtol=1E-5)
    np.testing.assert_allclose(sig_reload.axes_manager[2].scale, 0.0160616,
                    rtol=1E-5)


def test_default_header():
    # Simply check that no exceptions are raised
    header = get_default_header()
    assert header is not None


def test_non_square(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 6)
                                  ).astype(np.uint8))
    with pytest.raises(ValueError):
        signal.save(save_path, overwrite=True)


def test_load_lazy():
    from dask.array import Array
    s = hs.load(FILE2, lazy=True)
    assert isinstance(s.data, Array)


def test_load_to_memory():
    s = hs.load(FILE2, lazy=False)
    assert isinstance(s.data, np.ndarray)
    assert not isinstance(s.data, np.memmap)


def test_load_readonly():
    s = hs.load(FILE2, lazy=True)
    k = next(filter(lambda x: isinstance(x, str) and
                    x.startswith("array-original"),
                    s.data.dask.keys()))
    mm = s.data.dask[k]
    assert isinstance(mm, np.memmap)
    assert not mm.flags["WRITEABLE"]
    with pytest.raises(NotImplementedError):
        s.data[:] = 23


def test_load_inplace():
    with pytest.raises(ValueError):
        hs.load(FILE2, lazy=True, mmap_mode='r+')


def test_write_fresh(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)
                                  ).astype(np.uint8))
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)
    header = sarray2dict(get_default_header())
    header.update({
        'NX': 3, 'NY': 10,
        'DP_SZ': 5,
        'SX': 1, 'SY': 1,
        'SDP': 100,
        'Data_offset_2': 10 * 3 + header['Data_offset_1'],
        'Note': '',
    })
    header['Data_offset_2'] += header['Data_offset_2'] % 16
    assert (
        sig_reload.original_metadata.blockfile_header.as_dictionary() ==
        header)


def test_write_data_line(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(3, 5, 5)
                                  ).astype(np.uint8))
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)


def test_write_data_single(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(5, 5)
                                  ).astype(np.uint8))
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)


def test_write_data_am_mismatch(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)
                                  ).astype(np.uint8))
    signal.axes_manager.navigation_axes[1].size = 4
    with pytest.raises(ValueError):
        signal.save(save_path, overwrite=True)


def test_write_cutoff(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)
                                  ).astype(np.uint8))
    signal.axes_manager.navigation_axes[0].size = 20
    # Test that it raises a warning
    signal.save(save_path, overwrite=True)
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        sig_reload = hs.load(save_path)
        # There can be other warnings so >=
        assert len(w) >= 1
        warning_blockfile = ["Blockfile header" in str(warning.message)
                             for warning in w]
        assert True in warning_blockfile
        assert issubclass(w[warning_blockfile.index(True)].category,
                          UserWarning)
    cut_data = signal.data.flatten()
    pw = [(0, 17 * 10 * 5 * 5)]
    cut_data = np.pad(cut_data, pw, mode='constant')
    cut_data = cut_data.reshape((10, 20, 5, 5))
    np.testing.assert_equal(cut_data, sig_reload.data)


def test_crop_notes(save_path):
    note_len = 0x1000 - 0xF0
    note = 'test123' * 1000     # > note_len
    signal = hs.signals.Signal2D((255 * np.random.rand(2, 3, 2, 2)
                                  ).astype(np.uint8))
    signal.original_metadata.add_node('blockfile_header.Note')
    signal.original_metadata.blockfile_header.Note = note
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    assert (sig_reload.original_metadata.blockfile_header.Note ==
            note[:note_len])



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import copy
import itertools

import numpy as np
from hyperspy.drawing.marker import markers

class SpectrumLabelPosition():
    '''
    A class to get the positions of labels in spectrums such as EELS, EDX,
    XRF. The main method is the get_markers method which the user supplies a
    dictionary of edge labels specifying its energy positions, and it returns
    markers for labelling them.
    '''

    edge_label_style = {'ha' : 'center', 'va' : 'center',
                        'bbox' : dict(facecolor='white', alpha=0.2)}
    colour_list_label = ['black', 'darkblue', 'darkgreen',
                         'darkcyan', 'darkmagenta', 'dimgray',
                         'brown', 'deeppink', 'olive',
                         'crimson']

    def __init__(self, signal):
        self.signal = signal
        self.axis = self.signal.axes_manager.signal_axes[0]
        self._ele_col_dict = {}
        self.color_cycle = itertools.cycle(self.colour_list_label)

        self._set_active_figure_properties()

    def _set_active_figure_properties(self):
        # set the properties which depend on the figure
        self.signal_figure = self.signal._plot.signal_plot.figure
        self.figsize = self.signal_figure.get_size_inches()
        self.smin, self.smax = self.signal_figure.get_axes()[0].get_ylim()

        self.sig_index = self._get_current_signal_index()
        self.text_width, self.text_height = self._estimate_textbox_dimension()

    def _get_current_signal_index(self):
        # if it is a hyperspectrum, get the correct active figure
        if self.signal._plot.pointer is not None:
            sig_index = self.signal._plot.pointer.indices[0]
        else:
            sig_index = 0

        return sig_index

    def _check_signal_figure_changed(self):
        # check if the spectrum is changed
        # reset its properties if changed
        current_sig_index = self._get_current_signal_index()
        current_figsize = self.signal_figure.get_size_inches()
        if (current_sig_index != self.sig_index) or \
            not np.isclose(current_figsize, self.figsize).all():
            return True
        else:
            return False

    def _get_bbox_from_textbox_patch(self, fig, textbox):
        # get the bbox object of the textbox
        ax = fig.axes[0]
        r = fig.canvas.get_renderer()
        fig.draw(r)
        extent = textbox.get_bbox_patch().get_window_extent()
        bbox_patch = extent.transformed(ax.transData.inverted())

        return bbox_patch

    def _estimate_textbox_dimension(self, dummy_text='My_g8'):
        # get the dimension of a typical textbox in the current figure
        dummy_style = copy.deepcopy(self.edge_label_style)
        dummy_style['bbox']['alpha'] = 0
        dummy_style['alpha'] = 0
        tx = markers.text.Text(x=(self.axis.low_value+self.axis.high_value)/2,
                               y=(self.smin+self.smax)/2,
                               text=self._text_parser(dummy_text),
                               **dummy_style)
        self.signal.add_marker(tx)

        fig = tx.marker.get_figure()
        dummybb = self._get_bbox_from_textbox_patch(fig, tx.marker)
        tx.close()

        text_width = dummybb.width
        text_height = dummybb.height

        return text_width, text_height

    def get_markers(self, labels):
        '''Get the markers (vertical line segment and text box) for labelling
        the edges

        Parameters
        ----------
        labels : dictionary
            A dictionary with the labels as keys and their energies as values.
            E.g. for EELS edges it could be {'Mn_L2': 651.0, 'Cr_L3': 575.0}.

        Returns
        -------
        vls : list
            A list contains HyperSpy's vertical line segment marker
        txs : list
            A list contains HyperSpy's text marker
        '''

        xytext = self._get_textbox_pos(labels)

        vls = []
        txs = []
        for xyt in xytext:
            vl = markers.vertical_line_segment.VerticalLineSegment(x=xyt[0],
                                                                   y1=xyt[1],
                                                                   y2=xyt[2],
                                                                   color=xyt[4])
            tx = markers.text.Text(x=xyt[0], y=xyt[1],
                                   text=self._text_parser(xyt[3]), color=xyt[4],
                                   **self.edge_label_style)

            vl.events.closed.connect(self.signal._edge_marker_closed)
            tx.events.closed.connect(self.signal._edge_marker_closed)
            vl.auto_update = True
            tx.auto_update = True

            vls.append(vl)
            txs.append(tx)

        return vls, txs

    def _get_textbox_pos(self, edges, offset=None, step=None, lb=None,
                         ub=None):
        # get the information on placing the textbox and its properties
        if offset is None:
            offset = self.text_height
        if step is None:
            step = self.text_height
        if lb is None:
            lb = self.smin + offset
        if ub is None:
            ub = self.smax - offset

        if not self._ele_col_dict:
            self._ele_col_dict = self._element_colour_dict(edges)

        mid = (self.smax + self.smin) / 2
        itop = 1
        ibtm = 1

        xytext = []
        for edge in edges:
            energy = edges[edge]

            yval = self.signal.isig[float(energy)].data[self.sig_index]
            if yval <= mid: # from top
                y = ub - itop*step
                if y <= lb:
                    itop = 1
                    y = ub - itop*step
                itop += 1
            else: # from bottom
                y = lb + ibtm*step
                if y >= ub:
                    ibtm = 1
                    y = lb + ibtm*step
                ibtm += 1

            try:
                c = self._ele_col_dict[edge.split('_')[0]]
            except KeyError:
                self._ele_col_dict[edge.split('_')[0]] = next(self.color_cycle)
                c =  self._ele_col_dict[edge.split('_')[0]]

            xytext.append((energy, y, yval, edge, c))

        return xytext

    def _element_colour_dict(self, edges):
        # assign a colour to each element of the edges
        if isinstance(edges, dict):
            edges = edges.keys()

        elements = self._unique_element_of_edges(edges)

        d = {}
        for element in elements:
            d[element] = next(self.color_cycle)

        return d

    def _unique_element_of_edges(self, edges):
        # get the unique elements present in a sequence of edges
        elements = set()
        for edge in edges:
            element, _ = edge.split('_')
            elements.update([element])

        return elements

    def _text_parser(self, text_edge):
        # format the edge labels for LaTeX
        element, subshell = text_edge.split('_')

        if subshell[-1].isdigit():
            formatted = element+' '+'$\mathregular{'+subshell[0]+'_'+\
                subshell[-1]+'}$'
        else:
            formatted = element+' '+'$\mathregular{'+subshell[0]+'}$'

        return formatted



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of  GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np

from hyperspy._components.expression import Expression
from hyperspy._components.gaussian import _estimate_gaussian_parameters

sqrt2pi = math.sqrt(2 * math.pi)


tiny = np.finfo(np.float64).eps


class Doniach(Expression):

    r""" Doniach Sunjic lineshape

    .. math::
        :nowrap:

        \[
        f(x) = \frac{A \cos[ \frac{{\pi\alpha}}{2}+
        (1-\alpha)\tan^{-1}(\frac{x-centre+dx}{\sigma})]}
        {(\sigma^2 + (x-centre+dx)^2)^{\frac{(1-\alpha)}{2}}}
        \]


        \[
        dx = \frac{2.354820\sigma}{2 tan[\frac{\pi}{2-\alpha}]}
        \]


    =============== ===========
    Variable         Parameter
    =============== ===========
    :math:`A`        A
    :math:`\sigma`   sigma
    :math:`\alpha`   alpha
    :math:`centre`   centre
    =============== ===========

    Parameters
    -----------
    A : float
        Height
    sigma : float
        Variance parameter of the distribution
    alpha : float
        Tail or asymmetry parameter
    centre : float
        Location of the maximum (peak position).
    **kwargs
        Extra keyword arguments are passed to the ``Expression`` component.

    Note
    -----
    This is an asymmetric lineshape, originially design for xps but generally
    useful for fitting peaks with low side tails
    See Doniach S. and Sunjic M., J. Phys. 4C31, 285 (1970)
    or http://www.casaxps.com/help_manual/line_shapes.htm for a more detailed
    description

    """

    def __init__(self, centre=0., A=1., sigma=1., alpha=0.5,
                 module=["numpy", "scipy"], **kwargs):
        super(Doniach, self).__init__(
            expression="A*cos(0.5*pi*alpha+\
            ((1.0 - alpha) * arctan( (x-centre+offset)/sigma) ) )\
            /(sigma**2 + (x-centre+offset)**2)**(0.5 * (1.0 - alpha));\
            offset = 2.354820*sigma / (2 * tan(pi / (2 - alpha)))",
            name="Doniach",
            centre=centre,
            A=A,
            sigma=sigma,
            alpha=alpha,
            module=module,
            autodoc=False,
            **kwargs,
        )
        #
        self.sigma.bmin = 1.0e-8
        self.alpha.bmin = 1.0e-8
        self.isbackground = False
        self.convolved = True

    def estimate_parameters(self, signal, x1, x2, only_current=False):
        """Estimate the Donach by calculating the median (centre) and the
        variance parameter (sigma).

        Note that an insufficient range will affect the accuracy of this
        method and that this method doesn't estimate the asymmetry parameter
        (alpha).

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.

        only_current : bool
            If False estimates the parameters for the full dataset.

        Returns
        -------
        bool
            Returns True when the parameters estimation is successful

        Examples
        --------

        >>> g = hs.model.components1D.Lorentzian()
        >>> x = np.arange(-10, 10, 0.01)
        >>> data = np.zeros((32, 32, 2000))
        >>> data[:] = g.function(x).reshape((1, 1, 2000))
        >>> s = hs.signals.Signal1D(data)
        >>> s.axes_manager[-1].offset = -10
        >>> s.axes_manager[-1].scale = 0.01
        >>> g.estimate_parameters(s, -10, 10, False)
        """

        super()._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        centre, height, sigma = _estimate_gaussian_parameters(signal, x1, x2,
                                                              only_current)

        if only_current is True:
            self.centre.value = centre
            self.sigma.value = sigma
            self.A.value = height * 1.3
            if self.binned:
                self.A.value /= axis.scale
            return True
        else:
            if self.A.map is None:
                self._create_arrays()
            self.A.map['values'][:] = height * 1.3
            if self.binned:
                self.A.map['values'][:] /= axis.scale
            self.A.map['is_set'][:] = True
            self.sigma.map['values'][:] = sigma
            self.sigma.map['is_set'][:] = True
            self.centre.map['values'][:] = centre
            self.centre.map['is_set'][:] = True
            self.fetch_stored_values()
            return True



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy.drawing import image
from hyperspy.drawing.mpl_he import MPL_HyperExplorer
from hyperspy.defaults_parser import preferences


class MPL_HyperImage_Explorer(MPL_HyperExplorer):

    def plot_signal(self, **kwargs):
        """
        Parameters
        ----------
        **kwargs : dict
            The kwargs are passed to plot method of the image figure.

        """
        super().plot_signal()
        imf = image.ImagePlot()
        imf.axes_manager = self.axes_manager
        imf.data_function = self.signal_data_function
        imf.title = self.signal_title + " Signal"
        imf.xaxis, imf.yaxis = self.axes_manager.signal_axes

        # Set all kwargs value to the image figure before passing the rest
        # of the kwargs to plot method of the image figure
        for key, value in list(kwargs.items()):
            if hasattr(imf, key):
                setattr(imf, key, kwargs.pop(key))

        imf.quantity_label = self.quantity_label

        kwargs['data_function_kwargs'] = self.signal_data_function_kwargs
        if "cmap" not in kwargs.keys() or kwargs['cmap'] is None:
            kwargs["cmap"] = preferences.Plot.cmap_signal
        imf.plot(**kwargs)
        self.signal_plot = imf

        if imf.figure is not None:
            if self.axes_manager.navigation_axes:
                self.signal_plot.figure.canvas.mpl_connect(
                    'key_press_event', self.axes_manager.key_navigator)
            if self.navigator_plot is not None:
                self.navigator_plot.figure.canvas.mpl_connect(
                    'key_press_event', self.axes_manager.key_navigator)
                imf.events.closed.connect(self.close_navigator_plot, [])
            imf.events.closed.connect(self._on_signal_plot_closing, [])



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np

from hyperspy.components1d import Bleasdale


def test_function():
    g = Bleasdale()
    g.a.value = 1
    g.b.value = 2
    g.c.value = 2
    assert g.function(-0.5) == 0
    assert g.function(0) == 1
    assert g.function(12) == 0.2
    np.testing.assert_allclose(g.function(-.48),5)
    assert g.grad_a(0) == -0.5
    assert g.grad_b(0) == 0
    assert g.grad_c(0) == 0
    assert g.grad_a(-1) == 0
    assert g.grad_b(-1) == 0



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging

import traits.api as t

from hyperspy._signals.eds import (EDSSpectrum, LazyEDSSpectrum)
from hyperspy.defaults_parser import preferences
from hyperspy.ui_registry import add_gui_method, DISPLAY_DT, TOOLKIT_DT
from hyperspy.signal import BaseSetMetadataItems


_logger = logging.getLogger(__name__)


@add_gui_method(toolkey="hyperspy.microscope_parameters_EDS_SEM")
class EDSSEMParametersUI(BaseSetMetadataItems):

    beam_energy = t.Float(t.Undefined,
                          label='Beam energy (keV)')
    live_time = t.Float(t.Undefined,
                        label='Live time (s)')
    tilt_stage = t.Float(t.Undefined,
                         label='Stage tilt (degree)')
    azimuth_angle = t.Float(t.Undefined,
                            label='Azimuth angle (degree)')
    elevation_angle = t.Float(t.Undefined,
                              label='Elevation angle (degree)')
    energy_resolution_MnKa = t.Float(t.Undefined,
                                     label='Energy resolution MnKa (eV)')
    mapping = {
        'Acquisition_instrument.SEM.beam_energy': 'beam_energy',
        'Acquisition_instrument.TEM.Stage.tilt_alpha': 'tilt_stage',
        'Acquisition_instrument.SEM.Detector.EDS.live_time':
        'live_time',
        'Acquisition_instrument.SEM.Detector.EDS.azimuth_angle':
        'azimuth_angle',
        'Acquisition_instrument.SEM.Detector.EDS.elevation_angle':
        'elevation_angle',
        'Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa':
        'energy_resolution_MnKa', }


class EDSSEM_mixin:

    _signal_type = "EDS_SEM"

    def __init__(self, *args, **kwards):
        super().__init__(*args, **kwards)
        # Attributes defaults
        if 'Acquisition_instrument.SEM.Detector.EDS' not in self.metadata:
            if 'Acquisition_instrument.TEM' in self.metadata:
                self.metadata.set_item(
                    "Acquisition_instrument.SEM",
                    self.metadata.Acquisition_instrument.TEM)
                del self.metadata.Acquisition_instrument.TEM
        self._set_default_param()

    def get_calibration_from(self, ref, nb_pix=1):
        """Copy the calibration and all metadata of a reference.

        Primary use: To add a calibration to ripple file from INCA
        software

        Parameters
        ----------
        ref : signal
            The reference contains the calibration in its
            metadata
        nb_pix : int
            The live time (real time corrected from the "dead time")
            is divided by the number of pixel (spectrums), giving an
            average live time.

        Examples
        --------
        >>> ref = hs.datasets.example_signals.EDS_SEM_Spectrum()
        >>> s = hs.signals.EDSSEMSpectrum(
        >>>     hs.datasets.example_signals.EDS_SEM_Spectrum().data)
        >>> print(s.axes_manager[0].scale)
        >>> s.get_calibration_from(ref)
        >>> print(s.axes_manager[0].scale)
        1.0
        0.01

        """

        self.original_metadata = ref.original_metadata.deepcopy()
        # Setup the axes_manager
        ax_m = self.axes_manager.signal_axes[0]
        ax_ref = ref.axes_manager.signal_axes[0]
        ax_m.scale = ax_ref.scale
        ax_m.units = ax_ref.units
        ax_m.offset = ax_ref.offset

        # Setup metadata
        if 'Acquisition_instrument.SEM' in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.SEM
        elif 'Acquisition_instrument.TEM' in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.TEM
        else:
            raise ValueError(
                "The reference has no metadata.Acquisition_instrument.TEM"
                "\n nor metadata.Acquisition_instrument.SEM ")

        mp = self.metadata

        mp.Acquisition_instrument.SEM = mp_ref.deepcopy()

        if hasattr(mp_ref.Detector.EDS, 'live_time'):
            mp.Acquisition_instrument.SEM.Detector.EDS.live_time = \
                mp_ref.Detector.EDS.live_time / nb_pix

    def _load_from_TEM_param(self):
        """Transfer metadata.Acquisition_instrument.TEM to
        metadata.Acquisition_instrument.SEM

        """

        mp = self.metadata
        if mp.has_item('Acquisition_instrument.SEM') is False:
            mp.add_node('Acquisition_instrument.SEM')
        if mp.has_item('Acquisition_instrument.SEM.Detector.EDS') is False:
            mp.Acquisition_instrument.SEM.add_node('EDS')
        mp.Signal.signal_type = "EDS_SEM"

        # Transfer
        if 'Acquisition_instrument.TEM' in mp:
            mp.Acquisition_instrument.SEM = mp.Acquisition_instrument.TEM
            del mp.Acquisition_instrument.TEM

    def _set_default_param(self):
        """Set to value to default (defined in preferences)

        """
        mp = self.metadata
        if "Acquisition_instrument.SEM.Stage.tilt_alpha" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Stage.tilt_alpha",
                preferences.EDS.eds_tilt_stage)
        if "Acquisition_instrument.SEM.Detector.EDS.elevation_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.elevation_angle",
                preferences.EDS.eds_detector_elevation)
        if "Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa" \
                not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS."
                "energy_resolution_MnKa",
                preferences.EDS.eds_mn_ka)
        if "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle",
                preferences.EDS.eds_detector_azimuth)

    def set_microscope_parameters(self,
                                  beam_energy=None,
                                  live_time=None,
                                  tilt_stage=None,
                                  azimuth_angle=None,
                                  elevation_angle=None,
                                  energy_resolution_MnKa=None,
                                  display=True, toolkit=None):
        if set([beam_energy, live_time, tilt_stage, azimuth_angle,
                elevation_angle, energy_resolution_MnKa]) == {None}:
            tem_par = EDSSEMParametersUI(self)
            return tem_par.gui(toolkit=toolkit, display=display)
        md = self.metadata

        if beam_energy is not None:
            md.set_item("Acquisition_instrument.SEM.beam_energy", beam_energy)
        if live_time is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.live_time",
                live_time)
        if tilt_stage is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Stage.tilt_alpha",
                tilt_stage)
        if azimuth_angle is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle",
                azimuth_angle)
        if elevation_angle is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.elevation_angle",
                elevation_angle)
        if energy_resolution_MnKa is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS."
                "energy_resolution_MnKa",
                energy_resolution_MnKa)
    set_microscope_parameters.__doc__ = \
        """
        Set the microscope parameters.

        If no arguments are given, raises an interactive mode to fill
        the values.

        Parameters
        ----------
        beam_energy: float
            The energy of the electron beam in keV
        live_time : float
            In second
        tilt_stage : float
            In degree
        azimuth_angle : float
            In degree
        elevation_angle : float
            In degree
        energy_resolution_MnKa : float
            In eV
        {}
        {}

        Examples
        --------
        >>> s = hs.datasets.example_signals.EDS_SEM_Spectrum()
        >>> print('Default value %s eV' %
        >>>       s.metadata.Acquisition_instrument.
        >>>       SEM.Detector.EDS.energy_resolution_MnKa)
        >>> s.set_microscope_parameters(energy_resolution_MnKa=135.)
        >>> print('Now set to %s eV' %
        >>>       s.metadata.Acquisition_instrument.
        >>>       SEM.Detector.EDS.energy_resolution_MnKa)
        Default value 130.0 eV
        Now set to 135.0 eV

        """.format(DISPLAY_DT, TOOLKIT_DT)

    def _are_microscope_parameters_missing(self):
        """Check if the EDS parameters necessary for quantification
        are defined in metadata. If not, in interactive mode
        raises an UI item to fill the values

        """
        must_exist = (
            'Acquisition_instrument.SEM.beam_energy',
            'Acquisition_instrument.SEM.Detector.EDS.live_time', )

        missing_parameters = []
        for item in must_exist:
            exists = self.metadata.has_item(item)
            if exists is False:
                missing_parameters.append(item)
        if missing_parameters:
            _logger.info("Missing parameters {}".format(missing_parameters))
            return True
        else:
            return False

    def create_model(self, auto_background=True, auto_add_lines=True,
                     *args, **kwargs):
        """Create a model for the current SEM EDS data.

        Parameters
        ----------
        auto_background : boolean, default True
            If True, adds automatically a polynomial order 6 to the model,
            using the edsmodel.add_polynomial_background method.
        auto_add_lines : boolean, default True
            If True, automatically add Gaussians for all X-rays generated in
            the energy range by an element using the edsmodel.add_family_lines
            method.
        dictionary : {None, dict}, optional
            A dictionary to be used to recreate a model. Usually generated
            using :meth:`hyperspy.model.as_dictionary`

        Returns
        -------

        model : `EDSSEMModel` instance.

        """
        from hyperspy.models.edssemmodel import EDSSEMModel
        model = EDSSEMModel(self,
                            auto_background=auto_background,
                            auto_add_lines=auto_add_lines,
                            *args, **kwargs)
        return model


class EDSSEMSpectrum(EDSSEM_mixin, EDSSpectrum):
    pass


class LazyEDSSEMSpectrum(EDSSEMSpectrum, LazyEDSSpectrum):
    pass



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
import warnings
from distutils.version import LooseVersion

import numpy as np
import scipy
from scipy.linalg import svd
from scipy.sparse.linalg import svds

from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.misc.machine_learning.import_sklearn import (
    randomized_svd,
    sklearn_installed,
)

_logger = logging.getLogger(__name__)


def svd_flip_signs(u, v, u_based_decision=True):
    """Sign correction to ensure deterministic output from SVD.

    Adjusts the columns of u and the rows of v such that the loadings in the
    columns in u that are largest in absolute value are always positive.

    Parameters
    ----------
    u, v : numpy array
        u and v are the outputs of a singular value decomposition.
    u_based_decision : bool, default True
        If True, use the columns of u as the basis for sign flipping.
        Otherwise, use the rows of v. The choice of which variable to base the
        decision on is generally algorithm dependent.

    Returns
    -------
    u, v : numpy array
        Adjusted outputs with same dimensions as inputs.

    """
    # Derived from `sklearn.utils.extmath.svd_flip`.
    # Copyright (c) 2007-2020 The scikit-learn developers.
    # All rights reserved.

    if u_based_decision:
        max_abs_cols = np.argmax(np.abs(u), axis=0)
        signs = np.sign(u[max_abs_cols, range(u.shape[1])])
    else:
        max_abs_rows = np.argmax(np.abs(v), axis=1)
        signs = np.sign(v[range(v.shape[0]), max_abs_rows])

    u *= signs
    v *= signs[:, np.newaxis]

    return u, v


def svd_solve(
    data,
    output_dimension=None,
    svd_solver="auto",
    svd_flip=True,
    u_based_decision=True,
    **kwargs,
):
    """Apply singular value decomposition to input data.

    Parameters
    ----------
    data : numpy array, shape (m, n)
        Input data array
    output_dimension : None or int
        Number of components to keep/calculate
    svd_solver : {"auto", "full", "arpack", "randomized"}, default "auto"
        If auto:
            The solver is selected by a default policy based on `data.shape` and
            `output_dimension`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient "randomized"
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        If full:
            run exact SVD, calling the standard LAPACK solver via
            :py:func:`scipy.linalg.svd`, and select the components by postprocessing
        If arpack:
            use truncated SVD, calling ARPACK solver via
            :py:func:`scipy.sparse.linalg.svds`. It requires strictly
            `0 < output_dimension < min(data.shape)`
        If randomized:
            use truncated SVD, calling :py:func:`sklearn.utils.extmath.randomized_svd`
            to estimate a limited number of components
    svd_flip : bool, default True
        If True, adjusts the signs of the loadings and factors such that
        the loadings that are largest in absolute value are always positive.
        See :py:func:`~.learn.svd_pca.svd_flip` for more details.
    u_based_decision : bool, default True
        If True, and svd_flip is True, use the columns of u as the basis for sign-flipping.
        Otherwise, use the rows of v. The choice of which variable to base the
        decision on is generally algorithm dependent.

    Returns
    -------
    U, S, V : numpy array
        Output of SVD such that X = U*S*V.T

    """
    # Derived from `sklearn.decomposition.PCA`.
    # Copyright (c) 2007-2020 The scikit-learn developers.
    # All rights reserved.

    m, n = data.shape

    if output_dimension is None:
        output_dimension = min(m, n)
        if svd_solver == "arpack":
            output_dimension -= 1

    if svd_solver == "auto":
        if max(m, n) <= 500:
            svd_solver = "full"
        elif (
            output_dimension >= 1
            and output_dimension < 0.8 * min(m, n)
            and sklearn_installed
        ):
            svd_solver = "randomized"
        else:
            svd_solver = "full"

    if svd_solver == "randomized":
        if not sklearn_installed:  # pragma: no cover
            raise ImportError(
                "svd_solver='randomized' requires scikit-learn to be installed"
            )
        U, S, V = randomized_svd(data, n_components=output_dimension, **kwargs)
    elif svd_solver == "arpack":
        if LooseVersion(scipy.__version__) < LooseVersion("1.4.0"):  # pragma: no cover
            raise ValueError('`svd_solver="arpack"` requires scipy >= 1.4.0')

        if output_dimension >= min(m, n):
            raise ValueError(
                "svd_solver='arpack' requires output_dimension "
                "to be strictly less than min(data.shape)."
            )
        U, S, V = svds(data, k=output_dimension, **kwargs)
        # svds doesn't follow scipy.linalg.svd conventions,
        # so reverse its outputs
        S = S[::-1]
        # flip eigenvectors' sign to enforce deterministic output
        if svd_flip:
            U, V = svd_flip_signs(
                U[:, ::-1], V[::-1], u_based_decision=u_based_decision
            )
    elif svd_solver == "full":
        U, S, V = svd(data, full_matrices=False)
        # flip eigenvectors' sign to enforce deterministic output
        if svd_flip:
            U, V = svd_flip_signs(U, V, u_based_decision=u_based_decision)

        U = U[:, :output_dimension]
        S = S[:output_dimension]
        V = V[:output_dimension, :]

    return U, S, V


def svd_pca(
    data,
    output_dimension=None,
    svd_solver="auto",
    centre=None,
    auto_transpose=True,
    svd_flip=True,
    **kwargs,
):
    """Perform PCA using singular value decomposition (SVD).

    Read more in the :ref:`User Guide <mva.pca>`.

    Parameters
    ----------
    data : numpy array
        MxN array of input data (M features, N samples)
    output_dimension : None or int
        Number of components to keep/calculate
    svd_solver : {"auto", "full", "arpack", "randomized"}, default "auto"
        If auto:
            The solver is selected by a default policy based on `data.shape` and
            `output_dimension`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient "randomized"
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        If full:
            run exact SVD, calling the standard LAPACK solver via
            :py:func:`scipy.linalg.svd`, and select the components by postprocessing
        If arpack:
            use truncated SVD, calling ARPACK solver via
            :py:func:`scipy.sparse.linalg.svds`. It requires strictly
            `0 < output_dimension < min(data.shape)`
        If randomized:
            use truncated SVD, calling :py:func:`sklearn.utils.extmath.randomized_svd`
            to estimate a limited number of components
    centre : {None, "navigation", "signal"}, default None
        * If None, the data is not centered prior to decomposition.
        * If "navigation", the data is centered along the navigation axis.
        * If "signal", the data is centered along the signal axis.
    auto_transpose : bool, default True
        If True, automatically transposes the data to boost performance.
    svd_flip : bool, default True
        If True, adjusts the signs of the loadings and factors such that
        the loadings that are largest in absolute value are always positive.
        See :py:func:`~.learn.svd_pca.svd_flip` for more details.

    Returns
    -------
    factors : numpy array
    loadings : numpy array
    explained_variance : numpy array
    mean : numpy array or None (if centre is None)

    """
    N, M = data.shape

    if centre is None:
        mean = None

    else:
        # To avoid confusion between terminology in different
        # machine learning fields, we map the argument here.
        # See #1159 for some discussion.
        if centre in ["variables", "trials"]:
            centre_map = {
                "trials": "navigation",
                "variables": "signal",
            }
            centre_new = centre_map.get(centre, None)

            warnings.warn(
                f"centre='{centre}' has been deprecated and will be "
                f"removed in HyperSpy 2.0. Please use '{centre_new}' instead.",
                VisibleDeprecationWarning,
            )

            centre = centre_new

        if centre == "signal":
            mean = data.mean(axis=1)[:, np.newaxis]
        elif centre == "navigation":
            mean = data.mean(axis=0)[np.newaxis, :]
        else:
            raise ValueError("'centre' must be one of [None, 'navigation', 'signal']")

        data -= mean

    if auto_transpose is True:
        if N < M:
            _logger.info("Auto-transposing the data")
            data = data.T
        else:
            auto_transpose = False

    U, S, V = svd_solve(
        data,
        output_dimension=output_dimension,
        svd_solver=svd_solver,
        svd_flip=svd_flip,
        **kwargs,
    )

    explained_variance = S ** 2 / N

    if auto_transpose is False:
        factors = V.T
        loadings = U * S
    else:
        loadings = V.T
        factors = U * S

    return factors, loadings, explained_variance, mean
